# -*- coding: utf-8 -*-

import numpy

from simmate.toolkit import Structure

from pymatgen.analysis.diffusion.neb.pathfinder import (
    DistinctPathFinder,
    MigrationHop as PymatgenMigrationHop,
    IDPPSolver,
)

from typing import List


class MigrationImages(list):
    """
    This class is just a list of structures for a diffusion pathway. It has
    utility methods to help create these structures but otherwise behaves
    exactly like a python list.

    Note, this class is primarily used to generate inputs for calculations. If
    you'd like more advanced features, you should represent your diffusion
    pathway as a MigrationHop instead.As a rule of thumb: Only use this class
    if you are manually creating your pathway from endpoint supercells or from
    a set of supercell images.

    All MigrationHop's can be converted to MigrationImages (using the
    `from_migration_hop` method); but not all MigrationImages can be converted
    to MigrationHops.
    """

    def __init__(self, structures: List[Structure]):
        # This init function does nothing except apply typing -- specifically,
        # it says that it expects a list of structures.
        super().__init__(structures)

    def get_sum_structure(self, tolerance: float = 1e-3):
        """
        Takes all structures and combines them into one. Atoms that are within
        the given tolerance are joined into a single site.

        This is primarily used to view a diffusing pathway within a single
        structure -- as well as how the host lattice changes during diffusion.
        If you are able to convert your pathway to a MigrationHop, the
        MigrationHop.write_path() method is much faster and cleaner than this
        method, so it should be preffered. Also, because there are many atoms
        that are overlapping here, the output structure may cause programs
        like VESTA to crash.

        #### Parameters

        - `tolerance`:
            the angle and distance tolerance to consider fractional coordinates
            as matching. Matching sites will be merged as 1 site in the final
            sum structure.
        """

        # OPTIMIZE: this is very inefficient. It's much faster to visualize
        # structures with MigrationHop class because you know which atom is
        # moving. Here, we need to treat all atoms as moving. We can also
        # speed this up by only looking at diffusing species too.
        final_coords = []
        final_species = []
        for structure in self:  # recall self is a list of structures
            for site in structure:
                is_new = True
                for coords in final_coords:
                    if all(
                        numpy.isclose(
                            site.frac_coords,
                            coords,
                            rtol=tolerance,
                            atol=tolerance,
                        )
                    ):
                        is_new = False
                        break
                if is_new:
                    final_coords.append(site.frac_coords)
                    final_species.append(site.specie)

        structure = Structure(
            lattice=structure.lattice,
            species=final_species,
            coords=final_coords,
        )

        return structure

    @staticmethod
    def get_nimages(
        pathway_length: float,
        min_image_step: float = 0.7,
        require_midpoint: bool = True,
    ):
        """
        Gives the desirable number of images (not including start/end structures).

        This method helps generate a MigrationImages object, and typically is
        not called directly. The other classmethods of MigrationImages call
        this for you.

        #### Parameters

        - `pathway_length`:
            The length of the pathway.

        - `min_image_step`:
            The minimum step distance for the diffusing atom between images.
            The default is 0.7 Angstroms. For example, a path 2.8A long would
            require at least 4 images for this default.

        - `require_midpoint`:
            Whether there should be an image at the midpoint. In other words,
            whether the number of images should be odd. This is often important
            if you expect the transition state to be at the midpoint and you are
            not running CI-NEB. The default is True.

        Returns
        -------

        - `nimages`:
            The number of images to use for this pathway.

        """
        # At a minimum, we want to have images be 0.7 angstroms apart, and
        # with one additional image.
        nimages = pathway_length // min_image_step + 1

        # We also want an odd number of images. This ensures we have an image
        # at exactly the midpoint, which is often necessary if we aren't
        # running CI-NEB.
        if require_midpoint and nimages % 2 == 0:
            nimages += 1

        # This is a float but it makes more sense to have an integer
        return int(nimages)

    @classmethod
    def from_migration_hop(
        cls,
        migration_hop: PymatgenMigrationHop,
        vacancy_mode: bool = True,
        min_nsites: int = 80,
        max_nsites: int = 240,
        min_length: int = 10,
        **kwargs,
    ):
        """
        Creates a MigrationImages object from a MigrationHop object

        #### Parameters

        - `migration_hop`:
            The MigrationHop object that should be converted.

        - `vacancy_mode`:
            Whether to use single-vacancy diffusion (True) or interstitial
            diffusion (False). The default is True.

        - `min_nsites`:
            The minimum number of sites to have in the supercell structure.
            The default is 80.

        - `max_nsites`:
            The maximum number of sites to have in the supercell structure.
            The default is 240.

        - `min_length`:
            The minimum length for each vector in the supercell structure.
            The default is 10 Angstroms.

        - `**kwargs`:
            Any arguments that are normally accepted by IDPPSolver
        """
        # The third thing returned is the bulk_supercell which we don't need.
        start_supercell, end_supercell, _ = migration_hop.get_sc_structures(
            vac_mode=vacancy_mode,
            min_atoms=min_nsites,
            max_atoms=max_nsites,
            min_length=min_length,
        )

        # calculate the number of images required
        nimages = cls.get_nimages(migration_hop.length)

        return cls.from_endpoints(
            start_supercell,
            end_supercell,
            nimages=nimages,
            **kwargs,
        )

    @classmethod
    def from_endpoints(
        cls,
        structure_start: Structure,
        structure_end: Structure,
        nimages: int,
        **kwargs,
    ):
        """
        Creates a MigrationImages object from start and end supercell structures.
        You do not need to specify the diffusing atom(s) as all sites are
        linearly interpolated and then relaxed by IDPP.

        #### Parameters

        - `structure_start`:
            The starting supercell of the diffusion pathway.

        - `structure_end`:
            The ending supercell of the diffusion pathway.

        - `nimages`:
            The number of desired images for the pathway. Note, if you know the
            pathway length of your path, you can use the `get_nimages` static
            method to get a logical number of images.

        - `**kwargs`:
            Any arguments that are normally accepted by IDPPSolver
        """

        # Run IDPP relaxation on the images before returning them
        idpp_solver = IDPPSolver.from_endpoints(
            [structure_start, structure_end],
            nimages=nimages,
            **kwargs,
        )
        images = idpp_solver.run()

        return cls(images)

    @classmethod
    def from_startend_sites(
        cls,
        structure: Structure,
        site_start: int,
        site_end: int,
        **kwargs,
    ):
        """
        Creates a MigrationImages object from a bulk structure and start/end
        periodic sites of the diffusing atom.

        For example, this would allow a diffusion pathway that goes from a site
        at (0,0,0) to (1,1,1). Thus, symmetry and periodic boundry conditions
        are considered.

        Note, this method just creates a MigrationHop and then uses the
        `from_migration_hop` method to make a MigrationImages object.

        #### Parameters

        - `structure`:
            The bulk crystal structure (NOT the supercell).

        - `site_start`:
            The starting periodic site for this pathway.

        - `site_end`:
            The end periodic site for this pathway.

        - `**kwargs`:
            Any arguments that are normally accepted by `from_migration_hop`.
        """
        # This information is all we need for a MigrationHop object
        pathway = PymatgenMigrationHop(site_start, site_end, structure)
        return cls.from_migration_hop(pathway, **kwargs)

    @classmethod
    def from_structure(
        cls,
        structure: Structure,
        migrating_specie: str,
        pathfinder_kwargs: dict = {},
        **kwargs,
    ):
        """
        Given a bulk crystal structure, this will find all symmetrically
        unique pathways and return them as list of MigrationImages objects.

        #### Parameters

        - `structure`:
            The bulk crystal structure (NOT the supercell).

        - `migrating_specie`:
            The identity of the diffusing ion (e.g. "Li" or "Li1+"). Note, only
            provide oxidation state if you are using an oxidation-state decorated
            structure.

        - `pathfinder_kwargs`:
            Any arguments that are normally accepted by DistinctPathFinder, but
            given as a dictionary. The default is {}.

        - `**kwargs`:
            Any arguments that are normally accepted by `from_migration_hop`.
        """
        # convert to the LLL reduced primitive cell to make it as cubic as possible
        structure_lll = structure.get_sanitized_structure()

        # Use pymatgen to find all the symmetrically unique pathways.
        # NOTE: This only finds pathways up until the structure is percolating.
        # If you are interested in longer pathways, then this script needs to
        # be adjusted by passing additional kwargs
        pathfinder = DistinctPathFinder(
            structure_lll,
            migrating_specie=migrating_specie,
            **pathfinder_kwargs,
        )
        pathways = pathfinder.get_paths()

        # Now go through each path and convert to a MigrationPath. We return
        # these as a list of paths.
        migration_paths = []
        for pathway in pathways:

            migration_path = cls.from_migration_hop(
                migration_hop=pathway,
                **kwargs,
            )
            migration_paths.append(migration_path)

        return migration_paths

    @classmethod
    def from_dynamic(cls, migration_images):
        """
        This is an experimental feature. The code here is a repurposing of
        Structre.from_dynamic so consider making a general class for
        from_dynamic methods.
        """
        is_from_past_calc = False

        # assume any list is in the MigrationHop format if there are more than
        # two structures (i.e. there is at least one midpoint image)
        if isinstance(migration_images, list) and len(migration_images) > 2:
            migration_images_cleaned = migration_images

        else:
            raise Exception(
                "Unknown format provided for migration_images input."
                f"The following was provided: {str(migration_images)} of type "
                f"{type(migration_images)}"
            )

        migration_images_cleaned.is_from_past_calc = is_from_past_calc

        return migration_images_cleaned

    def as_dict(self):
        return [s.as_dict() for s in self]



# -*- coding: utf-8 -*-

from simmate.calculators.vasp.tasks.nudged_elastic_band.base import (
    VaspNudgedElasticBandTask,
)
from simmate.calculators.vasp.tasks.relaxation import MITRelaxation


class MITNudgedElasticBand(VaspNudgedElasticBandTask, MITRelaxation):
    """
    This task is a reimplementation of pymatgen's
    [MITNEBSet](https://pymatgen.org/pymatgen.io.vasp.sets.html#pymatgen.io.vasp.sets.MITNEBSet).

    Runs a NEB relaxation on a list of structures (aka images) using MIT Project
    settings. The lattice remains fixed and symmetry is turned off for this
    relaxation.

    You shouldn't use this workflow directly, but instead use the higher-level
    NEB workflows (e.g. diffusion/neb_all_paths or diffusion/neb_from_endpoints),
    which call this workflow for you.
    """

    incar = MITRelaxation.incar.copy()
    incar.update(
        dict(
            IBRION=1,
            ISYM=0,
            LCHARG=False,
            IMAGES__auto=True,
        )
    )
    incar.pop("multiple_keywords__smart_ldau")



# -*- coding: utf-8 -*-

import os
import tempfile
import time

from django.shortcuts import render

from simmate.toolkit import Structure
from simmate.database.base_data_types import Spacegroup
from simmate.visualization.structure.blender import make_blender_structure
from simmate.configuration.django import settings
from simmate.utilities import get_directory
from simmate.website.core_components.base_api_view import SimmateAPIViewSet


class SymmetryViewSet(SimmateAPIViewSet):
    table = Spacegroup
    template_list = "core_components/symmetry.html"
    template_retrieve = "core_components/spacegroup.html"


def structure_viewer(request):

    # Grabs all data after the '?' in the URL
    query = request.GET.dict()

    # convert the query to a Structure object
    structure_string = query.get("structure_string", "")
    structure = Structure.from_database_string(structure_string)

    # We want to make a 3d structure file. We make this so it...
    #   1. has a random name that ends with ".glb"
    #   2. is a temporary file (deletes once finished)
    #   3. is located in the static root folder
    # Note we make this file in the static directory because we want to use
    # "load_static" to grab it in the template
    temp_dir = get_directory(
        settings.STATIC_ROOT
    )  # use get_dir to ensure folder exists
    temp_file = tempfile.NamedTemporaryFile(
        dir=temp_dir,
        suffix=".glb",
        # BUG (the fix is below): if delete=True, this sets up a race condition
        # of when the user recieves the file versus when it is deleted.
        delete=False,
    )
    temp_filname_base = os.path.basename(temp_file.name)

    # BUG FIX: Because we can make these temporary files automatically delete,
    # we need to prevent 3D files from building up over time and taking up
    # room on the server/disk. We therefore automatically delete any files that
    # are older than N seconds. This process can be slow, but it adds minimal
    # overhead to the blender-creation process.
    detete_old_3d_files()

    # Use blender to create a temporary glb file. Note, the output here
    # may be useful for debugging, but it isn't used at the moment.
    output = make_blender_structure(structure, filename=temp_file.name)

    # we pass the base name to the template so that it knows where the static
    # file is located (template assumes static directory)
    context = {"3d_structure_filename": temp_filname_base}
    template = "core_components/structure_viewer.html"
    return render(request, template, context)


def test_viewer(request):

    # grab cif filenames to test with
    from simmate.toolkit import base_data_types

    structure_dir = os.path.join(
        os.path.dirname(base_data_types.__file__),
        "test",
        "test_structures",
    )
    cif_filenames = [os.path.join(structure_dir, f) for f in os.listdir(structure_dir)]

    structure = Structure.from_file(cif_filenames[0])

    context = {"structure": structure}
    template = "core_components/test.html"
    return render(request, template, context)


def detete_old_3d_files(time_cutoff: float = 60):
    """
    Goes through the static directory and finds all "tmp***.glb" files that
    are older than a given time cutoff. Each of these files is then deleted.

    #### Parameters

    - `time_cutoff`:
        The time (in seconds) required to determine whether a file is old or not.
        The default is 60 seconds.
    """
    # load the full path to the desired directory
    directory = os.path.join(settings.DJANGO_DIRECTORY, "static")

    # grab all files/folders in this directory and then limit this list to those
    # that are...
    #   1. NOT folders
    #   2. start with "tmp"
    #   3. end with ".glb"
    #   3. haven't been modified for at least time_cutoff
    filenames = []
    for filename in os.listdir(directory):
        filename_full = os.path.join(directory, filename)
        if (
            not os.path.isdir(filename_full)
            and filename.startswith("tmp")
            and filename.endswith(".glb")
            and time.time() - os.path.getmtime(filename_full) > time_cutoff
        ):
            filenames.append(filename_full)

    # now go through this list and delete the files that met the criteria.
    # If the file fails to be deleted, we just ingore it and move on
    for filename in filenames:
        try:
            os.remove(filename)
        except:
            continue



# -*- coding: utf-8 -*-

from simmate.database.base_data_types import table_column, Structure


class CodStructure(Structure):
    """
    Crystal structures from the [COD](http://www.crystallography.net/cod/) database.

    Currently, this table only stores the strucure, plus comments on whether the
    sturcture is ordered or has implicit hydrogens.
    """

    class Meta:
        app_label = "third_parties"

    base_info = ["id", "structure_string", "is_ordered", "has_implicit_hydrogens"]
    source = "The Crystallography Open Database"
    source_doi = "https://doi.org/10.1107/S0021889809016690"
    remote_archive_link = "https://archives.simmate.org/CodStructure-2022-02-20.zip"

    # These fields overwrite the default Structure fields due to a bug.
    chemical_system = table_column.TextField()
    formula_full = table_column.TextField()
    formula_reduced = table_column.TextField()
    formula_anonymous = table_column.TextField()
    # BUG: We can't use CharField for the COD database because there are a number
    # of structures that have 20+ elements in them and are disordered. The disordered
    # aspect throws issues in pymatgen where formulas can be returned as long
    # floats (ex: Ca2.1234567N). Until this is fixed and cleaned up, I'll
    # need to use TextField instead of CharField for these fields.

    id = table_column.CharField(max_length=25, primary_key=True)
    """
    The id used to represent the structure (ex: "cod-12345")
    """

    is_ordered = table_column.BooleanField()
    """
    whether the structure contains disordered sites (i.e. mixed occupancies)
    """

    has_implicit_hydrogens = table_column.BooleanField()
    """
    whether the structure has implicit Hydrogens. This means there should be
    Hydrogens in the structure, but they weren't explicitly drawn. Note,
    implicit hydrogens will make the chemical system and formula misleading 
    because of the absence of hydrogens.
    """

    @property
    def external_link(self) -> str:
        """
        URL to this structure in the COD website.
        """
        # All COD structures have their data mapped to a URL in the same way
        # ex: http://www.crystallography.net/cod/12345.html"
        # we store the id as "cod-123" so we need to convert this to uppercase
        id_formatted = self.id.split("-")[-1]
        return f"http://www.crystallography.net/cod/{id_formatted}.html"



# -*- coding: utf-8 -*-

"""
This file is important as your project grows and starts getting users.
Here, you need to make sure any changes you make don't break old features and
code.

Simmate does not have full guides on writing unittests yet, so we recommend
looking through other packges such as...
- [pytest](https://docs.pytest.org/en/7.1.x/)
- [pytest-django](https://pytest-django.readthedocs.io/en/latest/)
- [django test tutorials](https://docs.djangoproject.com/en/4.0/intro/tutorial05/)
"""

from django.test import TestCase

# Create your tests here.



# -*- coding: utf-8 -*-

from simmate.calculators.vasp.tasks.static_energy.materials_project import (
    MatprojStaticEnergy,
)


class MatprojPreBaderELF(MatprojStaticEnergy):
    """
    Runs a static energy calculation with a high-density FFT grid under settings
    from the Materials Project. Results can be used for Bader analysis where
    the ELF is used as the reference instead of the CHGCAR.
    """

    # The key thing for bader analysis is that we need a very fine FFT mesh. Other
    # than that, it's the same as a static energy calculation.
    incar = MatprojStaticEnergy.incar.copy()
    incar.update(
        LAECHG=True,  # write core charge density to AECCAR0 and valence to AECCAR2
        LELF=True,  # writes ELFCAR
        NPAR=1,  # must be set if LELF is set to True
        PREC="Single",  # ensures CHGCAR grid matches ELFCAR grid
        # Note that these set the FFT grid while the pre-Bader task sets the
        # fine FFT grid (e.g. useds NGX instead of NGXF)
        NGX__density_a=12,
        NGY__density_b=12,
        NGZ__density_c=12,
    )



# -*- coding: utf-8 -*-

# For adding new filters to django, look at...
# https://docs.djangoproject.com/en/3.2/howto/custom-template-tags/

import urllib

from django import template
from django.utils.safestring import mark_safe

from simmate.database.base_data_types import Structure as DatabaseStructure
from simmate.toolkit import Structure as ToolkitStructure

# We need a registration instance in order to configure everything with Django
register = template.Library()


@register.filter(name="structure_url")
def structure_to_url(structure):
    """
    Converts a toolkit Structure to a URL GET query. For example, a structure
    would return "?structure_string='...'". This is useful for when we want
    to pass a crystal structure directly from a URL -- which is what is
    done for our crystal structure viewer.
    """

    # if we are given a structure that is a database table instance, we simply
    # grab the structure string
    if isinstance(structure, DatabaseStructure):
        structure_string = structure.structure_string
    elif isinstance(structure, ToolkitStructure):
        # !!! This code should be located within a method on the toolkit class
        storage_format = "POSCAR" if structure.is_ordered else "CIF"
        structure_string = structure.to(fmt=storage_format)
        # !!!
    else:
        raise Exception("Unknown format provided.")

    # Take the output and convert it a URL query format
    url_query = urllib.parse.urlencode(dict(structure_string=structure_string))

    # Because we added new html to our script, we need to have Django check it
    # ensure it safe before returning. Read more about this here:
    # https://docs.djangoproject.com/en/3.2/howto/custom-template-tags/#filters-and-auto-escaping
    return mark_safe(url_query)



# -*- coding: utf-8 -*-

# I store all of my models elsewhere, so this file simply exists to show django where
# they are located at. I do this based on the directions given by:
# https://docs.djangoproject.com/en/3.1/topics/db/models/#organizing-models-in-a-package

# Rather than retyping all of the logic from this file, I use the copy method here.
from simmate.database.workflow_results import *



# -*- coding: utf-8 -*-

"""

> :warning: This file is only for use by the Simmate team. Users should instead
access data via the load_remote_archive method.

This file is for pulling AFLOW data into the Simmate database. 

AFLOW's supported REST API can be accessed via "AFLUX API". This is a separate
python package, which is maintained at https://github.com/rosenbrockc/aflow.
Note that this not from the official AFLOW team, but it is made such that keywords
are pulled dynamically from the AFLOW servers -- any updates in AFLOW's API should
be properly handled. Also structures are loaded as ASE Atom objects, which we then
convert to pymatgen.

"""

from tqdm import tqdm
from pymatgen.io.ase import AseAtomsAdaptor

from simmate.database.third_parties.aflow import AflowStructure


def load_all_structures():
    """
    Only use this function if you are part of the Simmate dev team!

    Loads all structures directly for the AFLOW database into the local
    Simmate database.
    """

    # AFLOW is not a dependency of simmate, so make sure you install it before using
    # this module
    try:
        from aflow import K as AflowKeywords
        from aflow.control import Query as AflowQuery
    except:
        raise ModuleNotFoundError(
            "You must install aflow client with `conda install -c conda-forge aflow`"
        )

    # The way we build a query looks similar to the Django API, where we start
    # with a Query object (similar to Table.objects manager) and build filters
    # off of it.
    data = (
        AflowQuery(
            # This is a list of the supported "catalogs" that AFLOW has -- which appear
            # to be separately stored databases. I just use all of them by default.
            catalog=[
                "icsd",  # 60,000 structures
                "lib1",  # 4,000 structures
                "lib2",  # 360,000 structures (binary phases)
                "lib3",  # 2,530,000 structures (ternary phases)
            ],
            # The batch size the number of results to return per HTTP request.
            batch_size=2000,
        )
        # .filter(
        #     # Now we want set the conditions for which structures to pull. Because we
        #     # want all of them, we normally comment this line out. For testing, we
        #     # can pull a smaller subset of the structures.
        #     # I use the element Dy because it gives about 1,300 structures
        #     AflowKeywords.species == "Dy",
        # )
        .select(
            # Indicate what data we want to grab from each result. Note that we don't
            # access the structure quite yet.
            AflowKeywords.auid,
            # This is the URL that leads to the rest of the data. Note it is a
            # interactive REST endpoint, while the dashboard link is different.
            # AflowKeywords.aurl,
            # The date that the entry was added
            # AflowKeywords.aflowlib_date,
            # Band gap
            # AflowKeywords.Egap,
            # The calculated energy of the unit cell
            AflowKeywords.enthalpy_cell,
            # BUG: or should we use energy_cell? Aren't these the same in
            # groundstate DFT?
        )
    )

    # Let's sanitize all structures first. So iterate through each one in the list
    # This also takes a while, so we use a progress bar (via tqdm)
    for entry in tqdm(data):

        # grab the structure -- this is loaded as an ASE atoms object
        structure_ase = entry.atoms()

        # convert the structure to pymatgen
        structure_pmg = AseAtomsAdaptor.get_structure(structure_ase)

        # now convert the entry to a database object
        structure_db = AflowStructure.from_toolkit(
            id=entry.auid.replace(":", "-"),
            structure=structure_pmg,
            energy=entry.enthalpy_cell,
        )

        # and save it to our database!
        structure_db.save()



# -*- coding: utf-8 -*-

from simmate.database.base_data_types import (
    DatabaseTable,
    table_column,
    Structure,
    Calculation,
    Forces,
    Thermodynamics,
    StaticEnergy,
    Relaxation,
    DynamicsRun,
)


class TestDatabaseTable(DatabaseTable):

    base_info = ["column1", "column2"]

    column1 = table_column.BooleanField()
    column2 = table_column.FloatField()

    class Meta:
        app_label = "test_app"


TestDatabaseTable2 = TestDatabaseTable.create_subclass(
    name="TestDatabaseTable2",
    module=__name__,  # required for serialization
    new_column3=table_column.BooleanField(),
)


class TestStructure(Structure):
    pass


class TestCalculation(Calculation):
    pass


class TestStructureCalculation(Structure, Calculation):
    pass


# Forces is always used with the Structure mix-in
class TestForces(Structure, Forces):
    pass


# We add the Structure mixin in order to test the "update stabilites" methods
class TestThermodynamics(Structure, Thermodynamics):
    pass


class TestStaticEnergy(StaticEnergy):
    pass


TestRelaxation, TestIonicStep = Relaxation.create_subclasses("Test", module=__name__)

TestDynamicsRun, TestDynamicsIonicStep = DynamicsRun.create_subclasses(
    "Test",
    module=__name__,
)



# -*- coding: utf-8 -*-

from dask.distributed import LocalCluster
import dask_jobqueue
from distributed.deploy.spec import SpecCluster


# These are the different Cluster types that you can import from dask_jobqueue
# TODO: Is there a way to dynamically determine this?
ALLOWED_CLUSTER_TYPES = [
    "Local",
    "PBS",
    "SLURM",
    "SGE",
    "HTCondor",
    "LSF",
    "Moab",
    "OAR",
]

HEADER_ART = r"""
 ____            _       ____ _           _
|  _ \  __ _ ___| | __  / ___| |_   _ ___| |_ ___ _ __
| | | |/ _` / __| |/ / | |   | | | | / __| __/ _ \ '__|
| |_| | (_| \__ \   <  | |___| | |_| \__ \ ||  __/ |
|____/ \__,_|___/_|\_\  \____|_|\__,_|___/\__\___|_|

"""


def run_cluster(
    cluster_type: str = "Local",
    n_workers: int = None,
    **cluster_kwargs,
) -> SpecCluster:

    # make sure we have a legal cluster type
    if cluster_type not in ALLOWED_CLUSTER_TYPES:
        raise Exception(
            f"{cluster_type} is not an allowed cluster type."
            f"Please choose one of the following: {ALLOWED_CLUSTER_TYPES}"
        )

    # Start a local cluster with default settings if requested
    if cluster_type == "Local":
        cluster = LocalCluster(
            n_workers=n_workers,
            preload="simmate.configuration.dask.connect_to_database",
            **cluster_kwargs,
        )

    # otherwise we have a jobqueue cluster
    else:
        # grab the class from the module. It will be named something like PBSCluster
        ClusterClass = getattr(dask_jobqueue, f"{cluster_type}Cluster")
        # Now initialize the cluster using the default Dask config.
        cluster = ClusterClass(**cluster_kwargs)
        if not n_workers:
            raise Exception("You must provide n_workers when using a JobQueue cluster")
        cluster.scale(n_workers)

    # print some info about the cluster for the user
    print(HEADER_ART)
    print(f"Scheduler is located at {cluster.scheduler.address}")
    print(f"Dashboard is located at {cluster.dashboard_link}")
    print(
        "To view the dashboard via ssh, use a command like... \n\t"
        "ssh -N -L 8787:example.net:8787 exampleuser@example.net"
    )
    # Also if we have a jobqueue, we can display the submit script being used.
    if issubclass(cluster.__class__, dask_jobqueue.JobQueueCluster):
        print("Workers are submitted to a jobqueue with the following script...\n")
        print(cluster.job_script())
    else:
        print("Workers are all ran on the local machine.")
    print("\n\n")

    return cluster



# -*- coding: utf-8 -*-

import pytest

from simmate.website.test_app.models import (
    TestDatabaseTable,
    TestDatabaseTable2,
)


@pytest.mark.django_db
def test_database_table():

    # add a row
    x = TestDatabaseTable(column1=True, column2=3.14)
    x.save()

    # check name
    assert x.get_table_name() == x.__class__.__name__


@pytest.mark.django_db
def test_create_subclass():
    # same as before but this table was made by the create_subclass method
    x = TestDatabaseTable2(column1=True, column2=3.14, new_column3=False)
    x.save()


@pytest.mark.django_db
def test_show_columns():
    TestDatabaseTable.show_columns()


@pytest.mark.django_db
def test_to_dataframe():
    TestDatabaseTable.objects.to_dataframe()


@pytest.mark.django_db
def test_to_toolkit():
    with pytest.raises(Exception):
        TestDatabaseTable.objects.to_toolkit()


@pytest.mark.django_db
def test_from_toolkit():

    # to object and save
    x = TestDatabaseTable.from_toolkit(column1=True, column2=3.14)
    x.save()

    # to a dictionary
    y = TestDatabaseTable.from_toolkit(column1=True, column2=3.14, as_dict=True)
    assert isinstance(y, dict)


@pytest.mark.django_db
def test_archive():

    # BUG: This test does not save archives within a tmpdir -- but instead the
    # working directory. Therefore, these tests cannot be split up and ran
    # in parallel (with pytest-xdist).

    # add sample rows
    x = TestDatabaseTable(column1=True, column2=3.14)
    x.save()
    y = TestDatabaseTable(column1=False, column2=-3.14)
    y.save()

    # Also try to load an archive that doesn't exist yet
    with pytest.raises(FileNotFoundError):
        TestDatabaseTable.load_archive(
            confirm_override=True,
        )

    # write to a file
    TestDatabaseTable.objects.to_archive()

    # try reloading without confirming override
    with pytest.raises(Exception):
        TestDatabaseTable.load_archive()

    # try again with confirmation. This is also our last test so we can
    # delete the archive when we're done
    TestDatabaseTable.load_archive(
        confirm_override=True,
        delete_on_completion=True,
    )

    # Our test table doesn't have the remote_archive_link label set.
    with pytest.raises(Exception):
        TestDatabaseTable.load_remote_archive(
            confirm_override=True,
        )

    # now add the attribute and try again
    # NOTE: This is a live CDN! If my CDN server goes down, this test will fail
    TestDatabaseTable.remote_archive_link = (
        "https://archives.simmate.org/TestDatabaseTable-2022-02-08.zip"
    )
    TestDatabaseTable.load_remote_archive(
        confirm_override=True,
    )



# -*- coding: utf-8 -*-

"""
This module defines the lowest-level classes for database tables and their
search results.

See the `simmate.database.base_data_types` (which is the parent module of 
this one) for example usage.
"""

import os
import inspect
import shutil
import urllib
import warnings

import yaml

import pandas
from django.db import models  # , transaction
from django_pandas.io import read_frame
from django.utils.timezone import datetime

from typing import List

# This line does NOTHING but rename a module. I have this because I want to use
# "table_column.CharField(...)" instead of "models.CharField(...)" in my Models.
# This let's beginners read my higher level classes and instantly understand what
# each thing represents -- without them needing to understand
# that Django Model == Database Table. Experts may find this annoying, so I'm
# sorry :(
from django.db import models as table_column


class SearchResults(models.QuerySet):
    """
    This class adds some extra methods to the results returned from a database
    search. For example, if you searched all Structures and wanted to convert
    these to a pandas dataframe or even a list of pymatgen structures, you can
    now do...

    ``` python
    # for a list of database objects (django models)
    search_results = Structures.objects.all()

    # for a pandas dataframe (which is like an Excel table)
    dataframe = search_results.to_dataframe()

    # for a list of pymatgen structure objects
    structures = search_results.to_toolkit()
    ```

    All other functionality is inherited from
    [Django QuerySets](https://docs.djangoproject.com/en/4.0/ref/models/querysets/).
    """

    def to_dataframe(
        self,
        fieldnames: List[str] = (),
        verbose: bool = True,
        index: str = None,
        coerce_float: str = False,
        datetime_index: str = False,
    ) -> pandas.DataFrame:
        """
        Returns a Pandas DataFrame of the search results

        This method is coppied from django_pandas'
        [manager.py](https://github.com/chrisdev/django-pandas/blob/master/django_pandas/managers.py)

        Paramaters
        -----------
        - `fieldnames`:
            The model field names(columns) to utilise in creating the DataFrame.
            You can span a relationships in the usual Django ORM way by using
            the foreign key field name separated by double underscores and refer
            to a field in a related model.
        - `index`:
            specify the field to use  for the index. If the index field is not
            in fieldnames it will be appended. This is mandatory for timeseries.
        - `verbose`:
            If  this is `True` then populate the DataFrame with the human
            readable versions for foreign key fields else use the actual values
            set in the model
        - `coerce_float`:
            Attempt to convert values to non-string, non-numeric objects (like
            decimal.Decimal) to floating point.
        - `datetime_index`: bool
            specify whether index should be converted to a DateTimeIndex.
        """

        # BUG: read_frame runs a NEW query, so it may be a different length from
        # the original queryset.
        # See https://github.com/chrisdev/django-pandas/issues/138 for issue
        return read_frame(
            self,
            fieldnames=fieldnames,
            verbose=verbose,
            index_col=index,
            coerce_float=coerce_float,
            datetime_index=datetime_index,
        )

    def to_toolkit(
        self,
    ) -> list:  # type of object varies (e.g. Structure, BandStructure, etc.)
        """
        Converts your SearchResults to a list of pymatgen objects
        """

        # This method will only be for structures and other classes that
        # support this method. So we make sure the model has supports it first.
        if not hasattr(self.model, "to_toolkit"):
            raise Exception(
                "This database table does not have a to_toolkit method implemented"
            )

        # now we can iterate through the queryset and return the converted
        # pymatgen objects as a list
        return [obj.to_toolkit() for obj in self]

    def to_archive(self, filename: str = None):
        """
        Writes a compressed zip file using the table's base_info attribute.
        Underneath, the file is written in a csv format.

        This is useful for small making archive files and reloading fixtures
        to a separate database.

        This method is attached to the table manager for scenarios to allow
        queryset filtering before dumping data.

        To load this database dump into a new database, use the class's
        `from_archive` method.

        #### Parameters

        - `filename`:
            The filename to write the zip file to. By defualt, None will make
            a filename named MyExampleTableName-2022-01-25.zip, where the date
            will be the current day (for versioning).
        """

        # Generate the file name if one wasn't given.
        if not filename:
            # This is automatically the name of the table plus the date, where
            # the date is for versioning. For example...
            #   MyExampleTable-2022-01-25
            today = datetime.today()
            filename_base = "-".join(
                [
                    self.model.__name__,
                    str(today.year),
                    str(today.month).zfill(2),
                    str(today.day).zfill(2),
                ]
            )
            filename = filename_base + ".zip"

        # Turn the filename into the full path. We do this because we only
        # want to
        # filename = os.path.abspath(filename)
        # os.path.dirname(filename)

        # grab the base_information, and if ID is not present, add it
        base_info = self.model.base_info
        if "id" not in base_info:
            base_info.append("id")
        # if "created_at" not in base_info:
        #     base_info.append("created_at")
        # if "updated_at" not in base_info:
        #     base_info.append("updated_at")

        # We want to load the entire table, but only grab the fields that
        # are in base_info.
        base_objs = self.only(*base_info)

        # now convert these objects to a pandas dataframe, again using just
        # the base_info columns
        df = base_objs.to_dataframe(fieldnames=base_info)

        # Write the data to a csv file
        # OPTIMIZE: is there a better format that pandas can write to?
        csv_filename = filename.replace(".zip", ".csv")
        df.to_csv(csv_filename, index=False)

        # now convert the dump file to a compressed zip. In the complex, os
        # functions below we are just grabbing the filename without the
        # zip ending. We are also grabbing the directory that the csv is
        # located in
        shutil.make_archive(
            base_name=filename.removesuffix(".zip"),
            format="zip",
            root_dir=os.path.dirname(os.path.abspath(csv_filename)),
            base_dir=os.path.basename(csv_filename),
        )

        # we can now delete the csv file
        os.remove(csv_filename)


# Copied this line from...
# https://github.com/chrisdev/django-pandas/blob/master/django_pandas/managers.py
# It simply converts this queryset class to a manager.
DatabaseTableManager = models.Manager.from_queryset(SearchResults)


class DatabaseTable(models.Model):
    """
    The base class for defining a table in the Simmate database. All tables and
    mixins inherit from this class.

    Usage is identical to
    [models in Django](https://docs.djangoproject.com/en/4.0/#the-model-layer)
    where this class only adds extra methods for convenience.
    """

    class Meta:
        abstract = True

    base_info: List[str] = []
    """
    The base information for this database table and only these fields are stored
    when the `to_archive` method is used. Using the columns in this list, all 
    other columns for this table can be calculated, so the columns in this list 
    are effectively the "raw data".
    """
    # TODO: setting base_info is becoming boilerplate. Consider making this
    # into a method like to_toolkit that automatically looks at all mix-ins
    # and creates this list (i.e. involves _base_info properties)

    created_at = table_column.DateTimeField(
        auto_now_add=True,
        blank=True,
        null=True,
    )
    """
    Timestamp of when this row was first added to the database table
    """

    updated_at = table_column.DateTimeField(
        auto_now=True,
        blank=True,
        null=True,
    )
    """
    Timestamp of when this row was was lasted changed / updated
    """

    source = table_column.JSONField(blank=True, null=True)
    """
    > Note, this field is highly experimental at the moment and subject to
    change.
    
    This column indicates where the data came from, and it could be a number 
    of things including...
     - a third party id
     - a structure from a different Simmate datbase table
     - a transformation of another structure
     - a creation method
     - a custom submission by the user
    
    By default, this is a JSON field to account for all scenarios, but some
    tables (such as those in `simmate.database.third_parties`) this is value
    should be the same for ALL entries in the table and therefore the column is
    overwritten as an attribute.
    
    For other tables where this is not a constant, here are some examples of
    values used in this column:
    
    ``` python
    # from a thirdparty database or separate table
    source = {
        "table": "MatprojStructure",
        "id": "mp-123",
    }
    
    # from a random structure creator
    source = {"method": "PyXtalStructure"}
    
    # from a templated structure creator (e.g. substituition or prototypes)
    source = {
        "method": "PrototypeStructure",
        "table": "AFLOWPrototypes",
        "id": 123,
    }
    
    # from a transformation
    source = {
        "method": "MirrorMutation",
        "table": "MatprojStructure",
        "id": "mp-123",
    }
    
    # from a multi-structure transformation
    source = {
        "method": "HereditaryMutation",
        "table": "MatprojStructure",
        "ids": ["mp-123", "mp-321"],
    }
    ```
    """
    # TODO: Explore polymorphic relations instead of a JSON dictionary.
    # Making relationships to different tables makes things difficult to use, so
    # these columns are just standalone.
    #
    # This is will be very important for "source" and "parent_nested_calculations"
    # fields because I have no way to efficiently convert these fields to the objects
    # that they refer to. There's also no good way to access a structure's "children"
    # (i.e. structure where they are the source).
    #
    # I should investigate generic relations in django though:
    # https://docs.djangoproject.com/en/3.2/ref/contrib/contenttypes/#generic-relations
    #
    # Another option is using django-polymorphic.
    # https://django-polymorphic.readthedocs.io/en/latest/
    # This thread is really helpful on the subject:
    # https://stackoverflow.com/questions/30343212/
    #
    # TODO: Consider adding some methods to track the history of a structure.
    #  This would be useful for things like evolutionary algorithms.
    # get_source_parent:
    #   this would iterate through sources until we find one in the same table
    #   as this one. Parent sources are often the most recent transformation
    #   or mutation applied to a structure, such as a MirrorMutation.
    # get_source_seed:
    #   this would iterate through sources until we hit a dead-end. So the seed
    #   source would be something like a third-party database, a method that
    #   randomly create structures, or a prototype.
    # Both of these get more complex when we consider transformation that have
    # multiple parents (and therefore multiple seeds too). An example of this
    # is the HereditaryMutation.

    source_doi: str = None
    """
    Source paper that must be referenced if this data is used. If this is None,
    please refer to the `source` attribute for further details on what to 
    reference.
    """

    remote_archive_link: str = None
    """
    The URL that is used to download the archive and then populate this table.
    Many tables, such as those in `simmate.database.third_parties`, have
    pre-existing data that you can download and load into your local database,
    so if this attribute is set, you can use the `load_remote_archive` method.
    """

    # I override the default manager with the one we define above, which has
    # extra methods useful for our querysets.
    objects = DatabaseTableManager()
    """
    Accesses all of the rows in this datatable and initiates a SearchResults
    object. Using this, you can perform complex filtering and conversions on
    data from this table.
    """

    @classmethod
    def get_table_name(cls) -> str:
        """
        Returns the name of this database table, which will simply match the
        class name. Using `Table.__name__` is often easier. This method simply
        makes the name accessible in Django Templates.
        """
        return cls.__name__

    @classmethod
    def create_subclass(
        cls,
        name: str,
        module: str,
        app_label: str = None,
        **new_columns,
    ):
        """
        This method is useful for dynamically creating a subclass DatabaseTable
        from some abstract class.

        Let's take an example where we inherit from a Structure table. The two
        ways we create a NewTable below are exactly the same:

        ``` python
        # Normal way to create a child class
        NewTable(Structure):
            new_field1 = table_column.FloatField()
            new_field2 = table_column.FloatField()

        # How this method makes the same child class
        NewTable = Structure.create_subclass(
            name="NewTable",
            module=__name__, # required for serialization
            new_field1 = table_column.FloatField()
            new_field2 = table_column.FloatField()
            # app_label --> typically not required bc the parent class sets this
        )
        ```

        While this might seem silly, it helps us avoid a bunch of boilerplate
        code when we need to redefine a bunch of relationships in every single
        child class (and always in the same way). A great example of it's utility
        is in `simmate.calculators.vasp.database.relaxation`.
        """

        # because we update values below, we make sure we are editting a
        # copy of the dictionary
        new_columns = new_columns.copy()

        # BUG: I'm honestly not sure what this does, but it works...
        # https://stackoverflow.com/questions/27112816/
        # For a more robust way of setting the Meta class, see my code in
        # simmate.website.workflows.utilities
        new_columns.update(
            {
                "__module__": module,
                "Meta": {"app_label": app_label} if app_label else {},
            }
        )
        # TODO: make it so I don't have to specify the module, but it is automatically
        # detected from where the class is created. This would remove boilerplate code.
        # A good start to this is here:
        #   https://stackoverflow.com/questions/59912684/
        # sys._getframe(1).f_globals["__name__"]  <-- grabs where this function is called
        # but doesn't work for multiple levels of inheritance. For example, this fails for
        # the Relaxation subclasses because the create_subclasses method calls
        # create_subclass -- therefore we'd need _getframe(3) instead of 1...

        # Now we dynamically create a new class that inherits from this main
        # one and also adds the new columns to it.
        NewClass = type(name, (cls,), new_columns)

        return NewClass

    @classmethod
    def from_toolkit(cls, as_dict: bool = False, **kwargs):
        """
        Given fundamental "base_info" and toolkit objects, this method will populate
        all relevant columns.

        If the table is made up of multiple mix-ins, this method iterates through
        each of the `_from_toolkit` methods of those mixins and combines the results.
        Therefore, you must view this method for each mix-ins to determine which
        kwargs must be passed.

        #### Parameters

        - `as_dict` :
            Whether to return the populated data as a dictionary or to initialize
            it as a database object. Defaults to False.
        - `**kwargs` :
            All fields required in order to initialize the database entry. For
            example, a Structure table would require a toolkit structure, while
            a Thermodynamics table would require an energy.

        Returns
        -------
        a dictionary if as_dict is True; and a database object if as_dict is False
        """

        # Grab a list of all parent classes
        parents = inspect.getmro(cls)

        # As we go through the parent classes below, we will identify the mixins
        # and populate data using those classes. All of this is fed in to our
        # main class at the end of the function. We keep this running dictionary
        # as we go.
        all_data = kwargs.copy()
        # TODO: How should I best handle passing extra kwargs to the final class
        # initialization? For example, I would want to pass `energy` but I wouldn't
        # want to pass the toolkit structure object. I may update this line in
        # the future to remove python objects. For now, I only remove structure
        # and migration_hop because I know that it is a toolkit object -- not
        # a database column.
        all_data.pop("structure", None)
        all_data.pop("migration_hop", None)
        all_data.pop("band_structure", None)
        all_data.pop("density_of_states", None)

        for parent in parents:

            # Skip the parent class if it doesn't directly inherit from the
            # DatabaseTable class. We do this because we want the fundamental
            # mixins that come with Simmate (such as Structure, Forces, Thermodynamics).
            # We also skip the classes that don't have a _from_toolkit method defined.
            # Right now, this is just the Calculation class that's skipped with this
            # condition.
            if parent.__base__ != DatabaseTable or not hasattr(parent, "_from_toolkit"):
                continue

            # Grab the input arguments for the _from_toolkit method. This will
            # give a list back like... ['cls', 'structure', 'as_dict']. We
            # don't consider as_dict or cls, so we remove those too
            inputs = inspect.getfullargspec(parent._from_toolkit).args
            inputs.remove("as_dict")
            inputs.remove("cls")

            # Now inputs is current of list of keys that we need. Next, we go
            # through our kwargs and see if we have any of these keys, and if so,
            # grab them.
            matching_inputs = {
                key: kwargs[key] for key in inputs if key in kwargs.keys()
            }

            # We now pass these inputs to the _from_toolkit. If we're missing
            # a required input, this will raise an error here. We only
            # want the expanded input, so we request a dictionary, not object.
            data = parent._from_toolkit(**matching_inputs, as_dict=True)

            # Now add this mixin's data to our collective dictionary
            all_data.update(data)

        # If as_dict is false, we build this into an Object. Otherwise, just
        # return the dictionary
        return all_data if as_dict else cls(**all_data)

    def update_from_toolkit(self, **kwargs):
        """
        Given fundamental "base_info" and toolkit objects, this method will populate
        all relevant columns.

        This method is meant for updating existing database entries with new
        data. If your creating a brand-new database entry, use the
        `from_toolkit` method instead.
        """
        new_kwargs = self.from_toolkit(as_dict=True, **kwargs)
        for new_kwarg, new_value in new_kwargs.items():
            setattr(self, new_kwarg, new_value)
        self.save()

    @classmethod
    def _confirm_override(
        cls,
        confirm_override: bool,
        parallel: bool,
        confirm_sqlite_parallel: bool,
    ):
        """
        A utility to make sure the user wants to load new data into their table
        and (if they are using sqlite) that they are aware of the risks of
        parallelizing their loading.

        This utility should not be called directly, as it is used within
        load_archive and load_remote_archive.
        """
        # first check if the table has data in it already. We raise errors
        # to stop the user from doing unneccessary and potentiall destructive
        # downloads
        if cls.objects.exists() and not confirm_override:
            # if the user has a third-party app, we can be more specific with
            # our error message.
            if cls._meta.app_label == "third_parties":
                raise Exception(
                    "It looks like you're using a third-party database table and "
                    "that the table already has data in it! This means you already "
                    "called load_remote_archive and don't need to do it again. "
                    "If you are trying reload a newer version of this data, make "
                    "sure you empty this table first. This can be done by "
                    "reseting your database or manually deleting all objects "
                    "with `ExampleTable.objects.all().delete()`"
                )

            # otherwise warning the user of overwriting data with matching
            # primary keys -- and ask them to use confirm_override.
            raise Exception(
                "It looks like this table already has data in it! By loading an "
                "archive, you could potentially overwrite this data. The most "
                "common mistake is non-unique primary keys between your current "
                "data and the archive -- if there is a duplicate primary key, it "
                "will overwrite your data. If you are confident the data is safe "
                "to load into your database, run this command again with "
                "confirm_override=True."
            )

        # Django and Dask can only handle so much for the parallelization
        # of database writing with SQLite. So if the user has SQLite as their
        # backend, we need to stop them from using this feature.
        from simmate.configuration.django.settings import DATABASES

        if parallel and not confirm_sqlite_parallel and "sqlite3" in str(DATABASES):
            raise Exception(
                "It looks like you are trying to run things in parallel but are "
                "using the default database backend (sqlite3), which is not "
                "always stable for massively parallel methods. You can still "
                "do this, but this message serves as a word of caution. "
                "You If you see error messages pop up saying 'database is "
                "locked', then your database is not stable at the rate you're "
                "trying to write data. This is a sign that you should either "
                "(1) switch to a different database backend such as Postgres "
                "or (2) reduce the parallelization of your tasks. If you are "
                "comfortable with these warnings and know what you're doing, "
                "set confirm_sqlite_parallel=True."
            )

    # @transaction.atomic  # We can't have an atomic transaction if we use Dask
    @classmethod
    def load_archive(
        cls,
        filename: str = None,
        delete_on_completion: bool = False,
        confirm_override: bool = False,
        parallel: bool = False,
        confirm_sqlite_parallel: bool = False,
    ):
        """
        Reads a compressed zip file made by `objects.to_archive` and loads the data
        back into the Simmate database.

        Typically, users won't call this method directly, but instead use the
        `load_remote_archive` method, which handles downloading the archive
        file from the Simmate website for you.

        #### Parameters

        - `filename`:
            The filename to write the zip file to. By defualt, None will try to
            find a file named "MyExampleTableName-2022-01-25.zip", where the date
            corresponds to version/timestamp. If multiple files match this format
            the most recent date will be used.

        - `delete_on_completion`:
            Whether to delete the archive file once all data is loaded into the
            database. Defaults to False

        - `confirm_override`:
            If the table already has data in it, the user must take particular
            care to downloading new data. This flag makes sure the user has
            made the proper checks to run this action. Default is False.

        - `parallel`:
            Whether to load the data in parallel. If true, this will start
            a local Dask cluster and each data row will be submitted as a task
            to the cluster. This provides substansial speed-ups for loading
            large datasets into the dataset. Default is False.

        - `confirm_sqlite_parallel`:
            If the database backend is sqlite, this parameter ensures the user
            knows what they are doing and know the risks of parallelization.
            Default is False.
        """

        # We disable warnings while loading archives because pymatgen prints
        # a lot of them (for things like rounding or electronegativity alerts).
        warnings.filterwarnings("ignore")

        # make sure the user actually wants to do this!
        cls._confirm_override(
            confirm_override,
            parallel,
            confirm_sqlite_parallel,
        )

        from tqdm import tqdm
        from simmate.toolkit import Structure as ToolkitStructure

        # generate the file name if one wasn't given
        if not filename:
            # The name will be something like "MyExampleTable-2022-01-25.zip".
            # We go through all files that match "MyExampleTable-*.zip" and then
            # grab the most recent date.
            matching_files = [
                file
                for file in os.listdir()
                if file.startswith(cls.__name__) and file.endswith(".zip")
            ]
            # make sure there is at least one file
            if not matching_files:
                raise FileNotFoundError(
                    f"No file found matching the {cls.__name__}-*.zip format"
                )
            # sort the files by date and grab the first
            matching_files.sort(reverse=True)
            filename = matching_files[0]

        # Turn the filename into the full path -- which makes a number of
        # manipulations easier below.
        filename = os.path.abspath(filename)

        # uncompress the zip file to the same directory that it is located in
        shutil.unpack_archive(
            filename,
            extract_dir=os.path.dirname(filename),
        )

        # We will now have a csv file of the same name, which we load into
        # a pandas dataframe
        csv_filename = filename.replace(".zip", ".csv")
        df = pandas.read_csv(csv_filename)

        # BUG: NaN values throw errors when read into SQL databases, so we
        # convert all NaN entries to None. This hacky line was taken from
        #   https://stackoverflow.com/questions/39279824/
        df = df.astype(object).where(df.notna(), None)

        # convert the dataframe to a list of dictionaries that we will iterate
        # through.
        entries = df.to_dict(orient="records")

        # to enable parallelization, we define a function to load a single
        # entry (or row) of data. This allows us to submit the function to Dask.
        def load_single_entry(entry):

            # For all entries, convert the structure_string to a toolkit structure
            if "structure_string" in entry:
                structure_str = entry.pop("structure_string")
                structure = ToolkitStructure.from_database_string(structure_str)
                entry["structure"] = structure
            # OPTIMIZE: is there a better way to do decide which entries need to be
            # converted to toolkit objects? This code may be better placed in
            # base_data_type methods (e.g. a `from_base_info` method for each)

            entry_db = cls.from_toolkit(**entry)
            entry_db.save()

        # now iterate through all entries to save them to the database
        if not parallel:
            # If user doesn't want parallelization, we run these in the main
            # thread and monitor progress with tqdm
            for entry in tqdm(entries):
                load_single_entry(entry)
        # otherwise we use dask to submit these in batches!
        else:

            from simmate.configuration.dask import batch_submit

            batch_submit(
                function=load_single_entry,
                args_list=entries,
                batch_size=15000,
            )

        # We can now delete the files. The zip file is only deleted if requested.
        os.remove(csv_filename)
        if delete_on_completion:
            os.remove(filename)  # the zip archive

    @classmethod
    def load_remote_archive(
        cls,
        remote_archive_link: str = None,
        confirm_override: bool = False,
        parallel: bool = False,
        confirm_sqlite_parallel: bool = False,
    ):
        """
        Downloads a compressed zip file made by `objects.to_archive` and loads
        the data back into the Simmate database.

        This method should only be called once -- when you have a completely
        empty database. After this call, all data will be stored locally and
        you don't need to call this method again (even accross python sessions).

        #### Parameters

        - `remote_archive_link`:
            The URL for that the archive will be downloaded from. If not supplied,
            it will default to the table's remote_archive_link attribute.

        - `confirm_override`:
            If the table already has data in it, the user must take particular
            care to downloading new data. This flag makes sure the user has
            made the proper checks to run this action.

        - `parallel`:
            Whether to load the data in parallel. If true, this will start
            a local Dask cluster and each data row will be submitted as a task
            to the cluster. This provides substansial speed-ups for loading
            large datasets into the dataset. Default is False.

        - `confirm_sqlite_parallel`:
            If the database backend is sqlite, this parameter ensures the user
            knows what they are doing and know the risks of parallelization.
            Default is False.
        """

        # make sure the user actually wants to do this!
        cls._confirm_override(
            confirm_override,
            parallel,
            confirm_sqlite_parallel,
        )

        # confirm that we have a link to download from
        if not remote_archive_link:
            # if no link was given we take the input value from class attribute
            if not cls.remote_archive_link:
                raise Exception(
                    "This table does not have a default link to load the archive "
                    " from. You must provide a remote_archive_link."
                )
            remote_archive_link = cls.remote_archive_link

        # tell the user where the data comes from
        if cls._meta.app_label == "third_parties":
            print(
                "\nWARNING: this data is NOT from the Simmate team, so be sure "
                "to visit the provider's website and to cite their work."
                f" This data is from {cls.source} and the following paper "
                f"should be cited: {cls.source_doi}\n"
            )

        # Predetermine the file name, which is just the ending of the URL
        archive_filename = remote_archive_link.split("/")[-1]

        # Download the archive zip file from the URL to the current working dir
        print("Downloading archive file...")
        urllib.request.urlretrieve(remote_archive_link, archive_filename)
        print("Done.\n")

        # now that the archive is downloaded, we can load it into our db
        print("Loading data into Simmate database...")
        cls.load_archive(
            archive_filename,
            delete_on_completion=True,
            confirm_override=True,  # we already confirmed this above
            parallel=parallel,
            confirm_sqlite_parallel=True,  # we already confirmed this above
        )
        print("Done.\n")

    @classmethod
    def get_column_names(cls) -> List[str]:
        """
        Returns a list of all the column names for this table and indicates which
        columns are related to other tables. This is primarily used to help
        view what data is available.
        """
        return [column.name for column in cls._meta.get_fields()]

    @classmethod
    def show_columns(cls):
        """
        Prints a list of all the column names for this table and indicates which
        columns are related to other tables. This is primarily used to help users
        interactively view what data is available.
        """
        # Iterate through and grab the columns. Note we don't use get_column_names
        # here because we are attaching relation data as well.
        column_names = [
            column.name + f" (relation to {column.related_model.__name__})"
            if column.is_relation
            else column.name
            for column in cls._meta.get_fields()
        ]

        # Then use yaml to make the printout pretty (no quotes and separate lines)
        print(yaml.dump(column_names))

    @classmethod
    def get_mixins(cls) -> list:  # -> List[DatabaseTable]
        """
        Grabs the mix-in Tables that were used to make this class. This will
        be mix-ins like Structure, Forces, etc. from the
        `simmate.database.base_data_types` module.
        """
        # this must be imported locally because it depends on all other classes
        # from this module -- and will create circular import issues if outside
        from simmate.database import base_data_types as simmate_mixins

        return [
            parent
            for parent in cls.__bases__
            if hasattr(simmate_mixins, parent.__name__)
            and parent.__name__ != "DatabaseTable"
        ]

    @classmethod
    def get_mixin_names(cls) -> List[str]:
        """
        Grabs the mix-in Tables that were used to make this class and returns
        a list of their names.
        """
        return [mixin.__name__ for mixin in cls.get_mixins()]

    @classmethod
    def get_extra_columns(cls) -> List[str]:
        """
        Finds all columns that aren't covered by the supported Table mix-ins.

        For example, a table made from...

        ``` python
        from simmate.database.base_data_types import (
            table_column,
            Structure,
            Forces,
        )

        class ExampleTable(Structure, Forces):
            custom_column1 = table_column.FloatField()
            custom_column2 = table_column.FloatField()
        ```

        ... would return ...

        ``` python
        ["custom_column1", "custom_column2"]
        ```
        """

        all_columns = cls.get_column_names()
        columns_w_mixin = [
            column for mixin in cls.get_mixins() for column in mixin.get_column_names()
        ]
        extra_columns = [
            column
            for column in all_columns
            if column not in columns_w_mixin and column != "id"
        ]
        return extra_columns



# -*- coding: utf-8 -*-

import os

from simmate.workflow_engine import ErrorHandler
from simmate.calculators.vasp.inputs import Incar


class LongVector(ErrorHandler):
    """
    This a simple error handler that is active when VASP finds an issue with the
    rotation matrix.
    """

    is_monitor = True
    filename_to_check = "vasp.out"
    possible_error_messages = [
        "One of the lattice vectors is very long (>50 A), but AMIN"
    ]

    def correct(self, directory: str) -> str:

        # load the INCAR file to view the current settings
        incar_filename = os.path.join(directory, "INCAR")
        incar = Incar.from_file(incar_filename)

        # make the fix
        incar["AMIN"] = 0.01
        correction = "switched AMIN to 0.01"

        # rewrite the INCAR with new settings
        incar.to_file(incar_filename)

        return correction



# -*- coding: utf-8 -*-

from django_filters import rest_framework as filters

from simmate.utilities import get_chemical_subsystems
from simmate.database.base_data_types import Structure as StructureTable


class Structure(filters.FilterSet):
    class Meta:
        model = StructureTable
        fields = dict(
            nsites=["range"],
            nelements=["range"],
            # elements=["contains"],
            density=["range"],
            density_atomic=["range"],
            volume=["range"],
            volume_molar=["range"],
            formula_full=["exact"],
            formula_reduced=["exact"],
            formula_anonymous=["exact"],
            spacegroup__number=["exact"],
            spacegroup__symbol=["exact"],
            spacegroup__crystal_system=["exact"],
            spacegroup__point_group=["exact"],
        )

    include_subsystems = filters.BooleanFilter(
        field_name="include_subsystems",
        label="Include chemical subsystems in results?",
        method="skip_filter",
    )
    """
    Whether to include subsystems of the given `chemical_system`. For example,
    the subsystems of Y-C-F would be Y, C, F, Y-C, Y-F, etc..
    """

    # include_suprasystems = forms.BooleanField(label="Include Subsytems", required=False)
    # TODO: Supra-systems would include all the elements listed AND more. For example,
    # searching Y-C-F would also return Y-C-F-Br, Y-Sc-C-F, etc.

    chemical_system = filters.CharFilter(method="filter_chemical_system")
    """
    The chemical system of the structure (e.g. "Y-C-F" or "Na-Cl")
    """

    def skip_filter(self, queryset, name, value):
        """
        For filter fields that use this method, nothing is done to queryset. This
        is because the filter is typically used within another field. For example,
        the `include_subsystems` field is not applied to the queryset, but it
        is used within the `filter_chemical_system` method.
        """
        return queryset

    def filter_chemical_system(self, queryset, name, value):
        # name/value here are the key/value pair for chemical system

        # Grab the "include_subsystems" field from the filter form. Note, this
        # value will be given as a string which we convert to a python boolean
        include_subsystems = self.data.dict().get("include_subsystems", "false")
        include_subsystems = True if include_subsystems == "true" else False

        # TODO:
        # Make sure that the chemical system is made of valid elements and
        # separated by hyphens

        # check if the user wants subsystems included (This will be True or False)
        if include_subsystems:
            systems_cleaned = get_chemical_subsystems(value)

        # otherwise just clean the single system
        else:
            # Convert the system to a list of elements
            systems_cleaned = value.split("-")
            # now recombine the list back into alphabetical order
            systems_cleaned = ["-".join(sorted(systems_cleaned))]
            # NOTE: we call this "systems_cleaned" and put it in a list so
            # that our other methods don't have to deal with multiple cases
            # when running a django filter.

        filtered_queryset = queryset.filter(chemical_system__in=systems_cleaned)

        # now return the cleaned value. Note that this is now a list of
        # chemical systems, where all elements are in alphabetical order.
        return filtered_queryset



# -*- coding: utf-8 -*-

from django.shortcuts import render, redirect
from django.contrib.auth.decorators import login_required

from simmate.database.base_data_types import DatabaseTable
from simmate import workflows as workflow_module
from simmate.workflows.utilities import (
    # WORKFLOW_TYPES,
    get_workflow,
    get_list_of_workflows_by_type,
)
from simmate.website.workflows.forms import SubmitWorkflow
from simmate.website.core_components.base_api_view import SimmateAPIViewSet


def workflows_all(request):

    # TODO: maybe instead load these descriptions from the simmate.{module}'s docstr
    # This would look something like...
    # all_metadata = {}
    # for flow_type in WORKFLOW_TYPES:
    #     --> grab the module
    #     --> use the __doc__ as the text.

    workflows_metadata = {
        "static-energy": (
            "These workflows calculate the energy for a structure. In many cases, "
            "this also involves calculating the lattice strain and forces for each site."
        ),
        "relaxation": (
            "These workflows geometry-optimize a structure's the lattice and sites "
            "until specified covergence criteria are met."
        ),
        "population-analysis": (
            "These workflows evaluate where electrons exist in a structure -- and "
            "they are often assigned to a specific site/atom. As a result of assigning "
            "electrons, these workflows give predicted oxidation states for each site."
        ),
        "band-structure": (
            "These workflows calculate the electronic band structure for a material."
        ),
        "density-of-states": (
            "These workflows calculate the electronic density of states for a material."
        ),
        "dynamics": (
            "These workflows run a molecular dynamics simulation for a material. This "
            "often involves iteratively evaluating the energy/forces of structure at "
            "specific temperature (or temperature ramp)."
        ),
        "diffusion": (
            "These workflows evaluate the diffusion of an atom through a material. "
            "At this time, these workflows are entirely Nudged-Elastic-Band (NEB) "
            "calculations."
        ),
    }

    # now let's put the data and template together to send the user
    context = {"workflows_metadata": workflows_metadata}
    template = "workflows/all.html"
    return render(request, template, context)


def workflows_by_type(request, workflow_type):

    workflow_names = get_list_of_workflows_by_type(workflow_type)

    workflow_type_module = getattr(workflow_module, workflow_type.replace("-", "_"))

    # pull the information together for each individual flow
    workflows = [get_workflow(n) for n in workflow_names]

    # now let's put the data and template together to send the user
    context = {
        "workflow_type": workflow_type,
        "workflow_type_description_short": workflow_type_module.__doc__,
        "workflows": workflows,
    }
    template = "workflows/by_type.html"
    return render(request, template, context)


class WorkflowAPIViewSet(SimmateAPIViewSet):

    template_list = "workflows/detail.html"
    template_retrieve = "workflows/detail_run.html"

    @classmethod
    def get_table(
        cls,
        request,
        workflow_type,
        workflow_name,
        pk=None,
    ) -> DatabaseTable:
        """
        grabs the relevant database table using the URL request
        """
        workflow_name_full = workflow_type + ".vasp." + workflow_name
        workflow = get_workflow(workflow_name_full)
        return workflow.database_table

    def get_list_context(
        self,
        request,
        workflow_type,
        workflow_name,
    ) -> dict:

        workflow_name_full = workflow_type + ".vasp." + workflow_name
        workflow = get_workflow(workflow_name_full)

        # TODO: grab some metadata about this calc. For example...
        # ncalculations = MITRelaxation.objects.count()
        # nflows_submitted = workflow.nflows_submitted

        return {
            "workflow": workflow,
            "flow_id": None,  # TODO
            "nflows_submitted": None,
        }

    def get_retrieve_context(
        self,
        request,
        workflow_type,
        workflow_name,
        pk,
    ) -> dict:

        workflow_name_full = workflow_type + ".vasp." + workflow_name
        workflow = get_workflow(workflow_name_full)

        return {"workflow": workflow}


@login_required
def workflow_submit(
    request,
    workflow_type: str,
    workflow_name: str,
):

    workflow_name_full = workflow_type + ".vasp." + workflow_name
    workflow = get_workflow(workflow_name_full)

    # dynamically create the form for this workflow
    FormClass = SubmitWorkflow.from_workflow(workflow)

    if request.method == "POST":
        submission_form = FormClass(request.POST, request.FILES)

        # To help with debugging:
        # raise Exception(str(submission_form.errors))

        if submission_form.is_valid():

            # We can now submit the workflow for the structure.
            flow_run_id = workflow.run_cloud(
                wait_for_run=False,
                **submission_form.cleaned_data,
            )

            return redirect(f"https://cloud.prefect.io/simmate/flow-run/{flow_run_id}")
    else:
        submission_form = FormClass()
    # now let's put the data and template together to send the user
    context = {
        "workflow": workflow,
        "submission_form": submission_form,
    }
    template = "workflows/submit.html"
    return render(request, template, context)



# -*- coding: utf-8 -*-

# from prefect.backend.flow_run import FlowRunView

from simmate.database.base_data_types import DatabaseTable, table_column


class Calculation(DatabaseTable):
    """
    A calculation is the result of a single `workflow.run()` call. Thus, a
    calculation is synonmyous with a Prefect "Flow-Run". Becuase of this,
    every table that inherits from this Calculation class will be directly
    linked to a specific Workflow. You can access this workflow via the
    `workflow` attribute.
    """

    class Meta:
        abstract = True
        app_label = "workflows"

    base_info = [
        "directory",
        "prefect_flow_run_id",
        "created_at",
        "updated_at",
        "corrections",
    ]

    directory = table_column.CharField(
        max_length=250,
        blank=True,
        null=True,
    )
    """
    This is the folder that the workflow was ran in. It will be something
    like... "/path/to/simmate-task-abc123"
    """
    # TODO: maybe add host computer name too? (via platform.node())

    prefect_flow_run_id = table_column.CharField(
        max_length=50,
        blank=True,
        null=True,
        unique=True,
    )
    """
    If this was submitted through Prefect, there is a lot more metadata available
    for this calculation. Simmate does not store this data directly, but instead
    we store the flow-run-id so that the user may look this up in the Prefect
    database if they wish.
    An example id is... d8a785e1-e344-463a-bede-0e7b3da7bab6
    """

    corrections = table_column.JSONField(blank=True, null=True)
    """
    Simmate workflows often have ErrorHandlers that fix any issues while the
    calaculation ran. This often involves changing settings, so we store
    any of those changes here.
    """

    @property
    def prefect_cloud_link(self) -> str:
        """
        URL to this calculation (flow-run) in the Prefect Cloud website.

        This assumes that the calculation was registered with prefect cloud and
        doesn't check to confirm it's been registered. To actually confirm that,
        use the `flow_run_view` attribute instead.
        """
        return f"https://cloud.prefect.io/simmate/flow-run/{self.prefect_flow_run_id}"

    @property
    def flow_run_view(self):  # -> FlowRunView
        """
        Checks if the prefect_flow_run_id was registered with Prefect Cloud, and
        if so, returns a
        [FlowRunView](https://docs.prefect.io/orchestration/flow-runs/inspection.html)
        that hold metadata such as the status. This metadata includes...
            - agent_id
            - auto_scheduled
            - context
            - created_by_user_id
            - end_time
            - flow_id
            - labels
            - name
            - parameters
            - scheduled_start_time
            - start_time
            - state
            - tenant_id
            - times_resurrected
            - version

        If Prefect Cloud is not configured or if the calculation was ran
        locally, the None is returned.
        """
        raise NotImplementedError("Migrating to Prefect 2.0")
        try:
            return FlowRunView.from_flow_run_id(self.prefect_flow_run_id)
        except:  # may fail if this is a local run or prefect api key not configured
            return None

    @property
    def prefect_flow_run_name(self) -> str:
        """
        Gives the user-friendly name of this run if the prefect_flow_run_id
        was registered with Prefect Cloud. (an example name is "friendly-bumblebee").
        """
        flowrunview = self.flow_run_view
        return flowrunview.name if flowrunview else None

    @property
    def prefect_state(self) -> str:
        """
        Gives the current state of this run if the prefect_flow_run_id
        was registered with Prefect Cloud. (ex: "Scheduled" or "Successful")
        """
        flowrunview = self.flow_run_view
        return flowrunview.state.__class__.__name__ if flowrunview else None

    @classmethod
    def from_prefect_id(cls, prefect_flow_run_id: str, **kwargs):
        """
        Given a prefect id, this method will do one of the following...

        1. if it exists, load the database object with matching prefect_id
        2. if it doesn't exist, create a new database object using this ID and extra kwargs

        It will then return the corresponding Calculation instance.
        """

        # Depending on how a workflow was submitted, there may be a calculation
        # extry existing already -- which we need to grab and then update. If it's
        # not there, we create a new one!

        # check if the calculation already exists in our database, and if so,
        # grab it and return it.
        if cls.objects.filter(prefect_flow_run_id=prefect_flow_run_id).exists():
            return cls.objects.get(prefect_flow_run_id=prefect_flow_run_id)
        # Otherwise we need to create a new one and return that.

        # To handle the initialization of other Simmate mix-ins, we pass all
        # information to the from_toolkit method rather than directly to cls.
        calculation = cls.from_toolkit(
            prefect_flow_run_id=prefect_flow_run_id,
            **kwargs,
        )
        calculation.save()

        return calculation

    # TODO: Consider adding resource use metadata (e.g. from VASP)
    # I may want to add these fields because Prefect doesn't store run stats
    # indefinitely AND it doesn't give detail memory use, number of cores, etc.
    # If all of these are too much, I could do a json field of "run_stats" instead
    #   - average_memory (The average memory used in kb)
    #   - max_memory (The maximum memory used in kb)
    #   - elapsed_time (The real time elapsed in seconds)
    #   - system_time(The system CPU time in seconds)
    #   - user_time(The user CPU time spent by VASP in seconds)
    #   - total_time (The total CPU time for this calculation)
    #   - cores (The number of cores used by VASP (some clusters print `mpi-ranks` here))

    # TODO: Consider linking to parent calculations for nested workflow runs
    # Where this calculation plays a role within a "nested" workflow calculation.
    # Becuase this structure can be reused by multiple workflows, we make this
    # a list of source-like objects. For example, a relaxation could be part of
    # a series of relaxations (like in StagedRelaxation) or it can be an initial
    # step of a BandStructure calculation.
    # parent_nested_calculations = table_column.JSONField(blank=True, null=True)



# -*- coding: utf-8 -*-

from simmate.workflow_engine import Workflow
from simmate.calculators.vasp.tasks.relaxation import (
    Quality04Relaxation as Quality04RelaxationTask,
)
from simmate.calculators.vasp.database.relaxation import (
    Quality04Relaxation as Quality04RelaxationResults,
)


class Relaxation__Vasp__Quality04(Workflow):
    s3task = Quality04RelaxationTask
    database_table = Quality04RelaxationResults
    description_doc_short = (
        "low-quality settings meant for highly unreasonable structures"
    )



# -*- coding: utf-8 -*-

"""
Workflows for relaxing a crystal structure
"""

# TODO: when I add more calculators, I can do something like this...
# if "simmate.calculators.Vasp" in installed_apps:
from simmate.calculators.vasp.workflows.relaxation.all import (
    Relaxation__Vasp__Matproj,
    Relaxation__Vasp__Mit,
    Relaxation__Vasp__NebEndpoint,
    Relaxation__Vasp__Quality00,
    Relaxation__Vasp__Quality01,
    Relaxation__Vasp__Quality02,
    Relaxation__Vasp__Quality03,
    Relaxation__Vasp__Quality04,
    Relaxation__Vasp__Staged,
)



# -*- coding: utf-8 -*-

import os

from simmate.calculators.vasp.inputs import Poscar


def test_poscar(tmpdir, structure):
    filename = os.path.join(tmpdir, "POSCAR")
    Poscar.to_file(structure, filename)
    structure_new = Poscar.from_file(filename)
    assert structure == structure_new



# -*- coding: utf-8 -*-

"""
WARNING: This module is experimental and subject to change.
"""

import os

from pymatgen.core.sites import PeriodicSite
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.analysis.transition_state import NEBAnalysis

from simmate.toolkit import Structure as ToolkitStructure
from simmate.toolkit.diffusion import MigrationHop as ToolkitMigrationHop
from simmate.database.base_data_types import (
    table_column,
    DatabaseTable,
    Structure,
)

# TODO: consider making a NestedCalculation
class DiffusionAnalysis(Structure):
    class Meta:
        abstract = True
        app_label = "workflows"

    migrating_specie = table_column.CharField(max_length=4, blank=True, null=True)
    """
    The element of the diffusion atom (e.g. "Li")
    """

    vacancy_mode = table_column.BooleanField(blank=True, null=True)
    """
    (if relevent) Whether vacancy or interstitial diffusion was used.
    """

    atomic_fraction = table_column.FloatField(blank=True, null=True)
    """
    Atomic fraction of the diffusion ion in the bulk structure.
    """

    # NOTE: barrier_cell, paths_involved, and npaths_involved need to be updated
    # by evaluating all MigrationHops.

    barrier_cell = table_column.FloatField(blank=True, null=True)
    """
    The energy barrier corresponding to the lowest-barrier percolation network
    """

    paths_involved = table_column.CharField(max_length=100, blank=True, null=True)
    """
    The symmetrically-unqiue pathways involved in the lowest-barrier 
    percolation network. This is given as a list of IDs.
    """
    # TODO: consider making this a relationship to MigrationPathways

    npaths_involved = table_column.IntegerField(blank=True, null=True)
    """
    The number of symmetrically-unqiue pathways involved in the lowest-barrier 
    percolation network
    """

    @classmethod
    def from_toolkit(
        cls,
        migrating_specie: str = None,
        vacancy_mode: bool = None,
        as_dict: bool = False,
        **kwargs,
    ):
        # the algorithm doesn't change for this method, but we do want to add
        # a few extra columns. Therefore we make the dictionary as normal and
        # then add those extra columns here.
        structure_dict = super().from_toolkit(as_dict=True, **kwargs)

        # now add the two extra columns
        structure_dict["migrating_specie"] = migrating_specie
        # Structure should be present -- otherwise an error would have been
        # raised by the call to super()
        structure_dict["atomic_fraction"] = (
            kwargs["structure"].composition.get_atomic_fraction(migrating_specie)
            if migrating_specie
            else None
        )

        # If as_dict is false, we build this into an Object. Otherwise, just
        # return the dictionary
        return structure_dict if as_dict else cls(**structure_dict)

    @classmethod
    def from_directory(cls, directory: str, **kwargs):
        """
        Creates a new database entry from a directory that holds diffusion analysis
        results. For now, this assumes the directory holds vasp output files.
        """

        # I assume the directory is from a vasp calculation, but I need to update
        # this when I begin adding new calculators.

        # TODO: It there a way I can figure out which tables the other calculations
        # are linked to? Specifically, the bulk relaxation, bulk static energy,
        # and the supercell relaxations. They are all in these directories too
        # but I can't save those results until I know where they belong.
        # Consider adding an attribute that points to those tables...? Or
        # maybe a relationship (which I'd rather avoid bc it makes things very
        # messy for users)
        # For now, I only grab the structure from the static-energy and store
        # it in the DiffusionAnalysis table.
        bulk_filename = os.path.join(
            directory,
            "static-energy.vasp.matproj",
            "POSCAR",
        )
        bulk_structure = ToolkitStructure.from_file(bulk_filename)

        # Save a diffusion analysis object so we can connect all other data
        # to it.
        analysis_db = cls.from_toolkit(
            structure=bulk_structure,
            vacancy_mode=True,  # assume this for now
        )
        analysis_db.save()

        # Iterate through all the subdirectories that start with "migration_hop*".
        # We also need to make sure we only grab directories because there are
        # also cifs present that match this naming convention.
        # We ignore the number when saving to avoid overwriting data.
        migration_directories = [
            os.path.join(directory, f)
            for f in os.listdir(directory)
            if os.path.isdir(os.path.join(directory, f))
            and f.startswith("migration_hop_")
        ]

        # now save each migration hop present
        for migration_dir in migration_directories:
            cls.migration_hops.field.model.from_directory(
                directory=migration_dir,
                diffusion_analysis_id=analysis_db.id,
            )

        return analysis_db

    @classmethod
    def create_subclasses(
        cls,
        name: str,
        module: str,
        **extra_columns,
    ):
        """
        Dynamically creates a subclass of Relaxation as well as a separate IonicStep
        table for it. These tables are linked together.

        Example use:

        ``` python
        from simmate.database.base_data_types import Relaxation

        # note the odd formatting here is just because we are parsing three
        # outputs from this method to three variables.
        (
            ExampleDiffusionAnalysis,
            ExampleMigrationHop,
            ExampleMigrationImage,
        ) = DiffusionAnalysis.create_subclasses(
            "Example",
            module=__name__,
        )
        ```

        #### Parameters

        - `name` :
            The prefix name of the subclasses that are output. "Relaxation" and
            "IonicStep" will be attached to the end of this prefix.

        - `module` :
            name of the module this subclass should be associated with. Typically,
            you should pass __name__ to this.

        - `**extra_columns` :
            Additional columns to add to the table. The keyword will be the
            column name and the value should match django options
            (e.g. table_column.FloatField())

        #### Returns

        - `NewDiffusionAnalysisClass` :
            A subclass of DiffusionAnalysis.
        - `NewMigrationHopClass`:
            A subclass of MigrationHop.
        - `NewMigrationHopClass`:
            A subclass of MigrationHop.
        """

        # For convience, we add columns that point to the start and end structures
        NewDiffusionAnalysisClass = cls.create_subclass(
            f"{name}DiffusionAnalysis",
            module=module,
            **extra_columns,
        )

        (
            NewMigrationHopClass,
            NewMigrationImageClass,
        ) = MigrationHop.create_subclasses_from_diffusion_analysis(
            name,
            NewDiffusionAnalysisClass,
            module=module,
            **extra_columns,
        )

        # we now have a new child class and avoided writing some boilerplate code!
        return NewDiffusionAnalysisClass, NewMigrationHopClass, NewMigrationImageClass


# TODO: consider making a Calculation bc this is what the corrections/directory
# information should be attached to.
class MigrationHop(DatabaseTable):
    class Meta:
        abstract = True
        app_label = "workflows"

    base_info = [
        "site_start",
        "site_end",
        "index_start",
        "index_end",
        "number",
    ]

    # OPTIMIZE: site_start and site_end
    # Really, this is a list of float values, but I save it as a string.
    # For robustness, should I save cartesian coordinates and/or lattice as well?
    # Does the max length make sense here and below?
    # I could also store as JSON since it's a list of coords.

    site_start = table_column.CharField(max_length=100, blank=True, null=True)
    """
    The starting fractional coordinates of the diffusing site.
    """

    site_end = table_column.CharField(max_length=100, blank=True, null=True)
    """
    The ending fractional coordinates of the diffusing site.
    """

    # BUG: the init script for MigrationHop can't identify the site indexes
    # properly but they should be the same as before because it is a symmetrized
    # structure. Note that even if I'm wrong in some case -- this will have
    # no effect because iindex and eindex are only used in one portion of
    # the hash as well as for printing the __str__ of the object.

    index_start = table_column.IntegerField(blank=True, null=True)
    """
    The starting site index of the diffusing site.
    """

    index_end = table_column.IntegerField(blank=True, null=True)
    """
    The ending site index of the diffusing site.
    """

    number = table_column.IntegerField(blank=True, null=True)
    """
    The expected index in DistinctPathFinder.get_paths. The shortest path is index 0
    and they are all ordered by increasing length.
    """

    length = table_column.FloatField(blank=True, null=True)
    """
    The length/distance of the pathway from start to end. Note, this is 
    linear measurement -- i.e. is it the length of the linear interpolated path.
    """

    dimension_path = table_column.IntegerField(blank=True, null=True)
    """
    The pathway dimensionality using the Larson scoring parameter.
    """

    dimension_host_lattice = table_column.IntegerField(blank=True, null=True)
    """
    Ignoring all atoms of the diffusing species, this is the dimensionality
    of the remaining atoms of the lattice using the Larson scoring parameter.
    """

    energy_barrier = table_column.FloatField(blank=True, null=True)
    """
    The energy barrier in eV. This evaluates all images of the diffusion pathway
    and find the difference between the maximum and minimum image energies.
    """

    # TODO:
    # Distance of the pathway relative to the shortest pathway distance
    # in the structure using the formula: (D - Dmin)/Dmin
    # distance_rel_min = models.FloatField()

    """ Relationships """
    # Each MigrationHop corresponds to the full analysis of one Structure, which
    # can have many MigrationHop(s)
    # diffusion_analysis = table_column.ForeignKey(
    #     # DiffusionAnalysis,
    #     on_delete=table_column.PROTECT,
    #     related_name="migration_hops",
    # )

    # TODO:
    # image_start --> OneToOneField for specific MigrationHop
    # image_end --> OneToOneField for specific MigrationHop
    # image_transition_state --> OneToOneField for specific MigrationHop

    # Just like Relaxation points to IonicSteps, NEB will point to MigrationImages

    @classmethod
    def _from_toolkit(
        cls,
        migration_hop: ToolkitMigrationHop,
        as_dict: bool = False,
        number: int = None,
        **kwargs,
    ):

        # convert the pathway object into the database table format
        hop_dict = dict(
            site_start=" ".join(str(c) for c in migration_hop.isite.frac_coords),
            site_end=" ".join(str(c) for c in migration_hop.esite.frac_coords),
            index_start=migration_hop.iindex,
            index_end=migration_hop.eindex,
            length=migration_hop.length,
            # diffusion_analysis_id=diffusion_analysis_id,
            **kwargs,
        )

        # If as_dict is false, we build this into an Object. Otherwise, just
        # return the dictionary
        return hop_dict if as_dict else cls(**hop_dict)

    # BUG: because of rounding in the from_toolkit method, the get_sc_structures
    # is unable to identify equivalent sites. I opened an issue for this
    # with their team:
    #   https://github.com/materialsvirtuallab/pymatgen-analysis-diffusion/issues/296
    def to_toolkit(self) -> ToolkitMigrationHop:
        """
        converts the database MigrationHop to a toolkit MigrationHop
        """
        # The bulk crystal structure is stored in the diffusion analysis table
        structure = self.diffusion_analysis.to_toolkit()

        # pathways require a symmetrized structure
        # BUG: I need to ensure the symmetrized structure comes out the same
        # as before, so I need to be careful with updates and hardcoding here.
        sga = SpacegroupAnalyzer(structure, symprec=0.1)
        symm_structure = sga.get_symmetrized_structure()

        isite_new = PeriodicSite(
            species=self.diffusion_analysis.migrating_specie,
            coords=[float(x) for x in self.site_start.split(" ")],
            lattice=symm_structure.lattice,
        )
        esite_new = PeriodicSite(
            species=self.diffusion_analysis.migrating_specie,
            coords=[float(x) for x in self.site_end.split(" ")],
            lattice=symm_structure.lattice,
        )

        path = ToolkitMigrationHop(isite_new, esite_new, symm_structure)
        # BUG: see comment attached to definition of these fields
        path.iindex = self.index_start
        path.eindex = self.index_end

        # if the pathways match, then we can return the pathway object!
        return path

    #######
    # BUG: I need to distinguish between the from_toolkit/to_toolkit methods
    # above that just load a MigrationImages object vs. the from_directory and
    # from_pymatgen methods below that include results. As-is these similar
    # names but very different use cases makes things confusing for users.
    #######

    @classmethod
    def from_directory(cls, directory: str, **kwargs):
        # I assume the directory is from a vasp calculation, but I need to update
        # this when I begin adding new calculators.

        # BUG: A fix is make during the workup() method that may be relevant here.
        # simmate.calculators.vasp.tasks.nudged_elastic_band.MITNudgedElasticBand.workup
        analysis = NEBAnalysis.from_dir(directory)
        return cls.from_pymatgen(analysis=analysis, **kwargs)

    @classmethod
    def from_pymatgen(
        cls,
        analysis: NEBAnalysis,
        diffusion_analysis_id: int = None,  # only used if updating
        migration_hop_id: int = None,  # only used if updating
    ):
        # We reference these related tables frequently below so it easier to
        # grab them up front.
        diffusion_analysis_table = cls.diffusion_analysis.field.related_model
        migration_image_table = cls.migration_images.field.model

        # First, we need a migration_hop database object.
        # All of hops should link to a diffusion_analysis entry, so we check
        # for that here too. The key thing of these statements is that we
        # have a migration_hop_id at the end.

        # If no ids were given, then we make a new entries for each.
        if not diffusion_analysis_id and not migration_hop_id:
            # Note, we have minimal information if this is the case, so these
            # table entries will have a bunch of empty columns.

            # We don't have a bulk structure to use for this class, so we use
            # the first image
            analysis_db = diffusion_analysis_table.from_toolkit(
                structure=analysis.structures[0],
                vacancy_mode=True,  # assume this for now
            )
            analysis_db.save()

            # This table entry will actually be completely empty... It only
            # serves to link the MigrationImages together
            hop_db = cls(
                diffusion_analysis_id=analysis_db.id,
            )
            hop_db.save()
            migration_hop_id = hop_db.id

        elif diffusion_analysis_id and not migration_hop_id:
            # This table entry will actually be completely empty... It only
            # serves to link the MigrationImages together
            hop_db = cls(diffusion_analysis_id=diffusion_analysis_id)
            hop_db.save()
            migration_hop_id = hop_db.id

        elif migration_hop_id:
            # We don't use the hop_id, but making this query ensures it exists.
            hop_db = cls.objects.get(id=migration_hop_id)
            # Even though it's not required, we make sure the id given for the
            # diffusion analysis table matches this existing hop id.
            if diffusion_analysis_id:
                assert hop_db.diffusion_analysis.id == diffusion_analysis_id

        # Now same migration images and link them to this parent object.
        # Note, the start/end Migration images will exist already in the
        # relaxation database table. We still want to save them again here for
        # easy access.
        for image_number, image_data in enumerate(
            zip(
                analysis.structures,
                analysis.energies,
                analysis.forces,
                analysis.r,
            )
        ):
            image, energy, force, distance = image_data
            image_db = migration_image_table.from_toolkit(
                structure=image,
                number=image_number,
                force_tangent=force,
                energy=energy,
                structure_distance=distance,
                migration_hop_id=migration_hop_id,
            )
            image_db.save()

        return hop_db

    @classmethod
    def create_subclasses_from_diffusion_analysis(
        cls,
        name: str,
        diffusion_analysis: DiffusionAnalysis,
        module: str,
        **extra_columns,
    ):
        """
        Dynamically creates subclass of MigrationHop and MigrationImage, then
        links them to the DiffusionAnalysis table.

        This method should NOT be called directly because it is instead used by
        `DiffusionAnalysis.create_subclasses`.

        #### Parameters

        - `name` :
            Name of the subclass that is output.
        - `diffusion_analysis` :
            DiffusionAnalysis table that these subclasses should be associated with.
        - `module` :
            name of the module this subclass should be associated with. Typically,
            you should pass __name__ to this.
        - `**extra_columns` :
            Additional columns to add to the table. The keyword will be the
            column name and the value should match django options
            (e.g. table_column.FloatField())

        #### Returns

        - `NewMigrationHopClass`:
            A subclass of MigrationHop.
        - `NewMigrationHopClass`:
            A subclass of MigrationHop.
        """

        NewMigrationHopClass = cls.create_subclass(
            f"{name}MigrationHop",
            diffusion_analysis=table_column.ForeignKey(
                diffusion_analysis,
                on_delete=table_column.CASCADE,
                related_name="migration_hops",
            ),
            module=module,
            **extra_columns,
        )

        NewMigrationImageClass = MigrationImage.create_subclass_from_migration_hop(
            name,
            NewMigrationHopClass,
            module=module,
            **extra_columns,
        )

        return NewMigrationHopClass, NewMigrationImageClass


class MigrationImage(Structure):
    class Meta:
        abstract = True
        app_label = "workflows"

    base_info = [
        "number",
        "structure_string",
        "force_tangent",
        "structure_distance",
        "energy",
    ]

    number = table_column.IntegerField()
    """
    The image number. Note, the starting image is always 0 and the ending
    image is always -1. The choice of -1 is because we often add the ending
    image to the database table before knowing how many midpoint images we
    are going to create.
    """

    force_tangent = table_column.FloatField(blank=True, null=True)
    """
    The tangent force of image. Not to be confused with forces from the Forces mix-in.
    """

    # Note, we only care about the total energy for NEB images -- not the other
    # fields that the Thermodynamics mix-in provides. This is why we set this
    # field directly, rather than using the mix-in.
    energy = table_column.FloatField(blank=True, null=True)
    """
    The calculated total energy. Units are in eV. 
    """

    structure_distance = table_column.FloatField(blank=True, null=True)
    """
    The fingerprint distance of the image from the starting image. A smaller
    means the structures are more similar, with 0 being an exact match.
    """

    # We don't need the source column for the MigrationImage class because we
    # instead stored the source on the DiffusionAnalysis object. This line
    # deletes the source column from our Structure mix-in.
    source = None

    @classmethod
    def create_subclass_from_migration_hop(
        cls,
        name: str,
        migration_hop: MigrationHop,
        module: str,
        **extra_columns,
    ):
        """
        Dynamically creates subclass of MigrationImage, then links it to the
        MigrationHop table.

        This method should NOT be called directly because it is instead used by
        `DiffusionAnalysis.create_subclasses`.

        #### Parameters

        - `name` :
            Name of the subclass that is output.
        - `migration_hop` :
            MigrationHop table that these images should be associated with.
        - `module` :
            name of the module this subclass should be associated with. Typically,
            you should pass __name__ to this.
        - `**extra_columns` :
            Additional columns to add to the table. The keyword will be the
            column name and the value should match django options
            (e.g. table_column.FloatField())

        #### Returns

        - `NewClass` :
            A subclass of MigrationImage.

        """

        NewClass = cls.create_subclass(
            f"{name}MigrationImage",
            migration_hop=table_column.ForeignKey(
                migration_hop,
                on_delete=table_column.CASCADE,
                related_name="migration_images",
            ),
            module=module,
            **extra_columns,
        )

        return NewClass



# -*- coding: utf-8 -*-

from typing import List

import click
from click import Context


@click.group()
def workflows():
    """A group of commands for running workflows or viewing their settings."""
    pass


def list_options(options: List) -> int:
    """
    This is a utility for click (cli) that prints of list of items as a numbered
    list. It prompts users to select an option from the list.

    For example, `["item1", "item2", "item3"]` would print...
    ```
        (01) item1
        (02) item2
        (03) item3
    ```

    #### Parameters

    - `options`:
        a list of strings to choose from

    #### Returns

    - `selected_index`:
        The integer value of the choice selected. This will follw python indexing
        so the index of the options list. (e.g. if item1 was selected, 0 would be
        returned)

    """

    for i, item in enumerate(options):
        number = str(i + 1).zfill(2)
        click.echo(f"\t({number}) {item}")

    # Have the user select an option. We use -1 because indexing count is from 0, not 1.
    selected_index = click.prompt("\n\nPlease choose a number:", type=int) - 1

    if selected_index >= len(options) or selected_index < 0:
        raise click.ClickException(
            "Number does not match any the options provided. Exiting."
        )

    click.echo(f"You have selectd `{options[selected_index]}`.")

    return selected_index


def parse_parameters(context: Context) -> dict:
    """
    This is a utility for click (cli) that formats input parameters for workflow
    runs. It accounts for recieving a `click.Context` object and ensures we
    return a dictionary.

    In order to provide a context, make sure the click command has the following:

    ``` python
    @click.command(
        context_settings=dict(
            ignore_unknown_options=True,
            allow_extra_args=True,
        )
    )
    @click.pass_context
    def example(context):
        pass
    ```

    These context settings let us pass extra kwargs, as some workflows have unique
    inputs (e.g. structures, migration_hop, etc. instead of a structure input).
    This is also why we have pass_context and then the kwarg context.
    """

    # The extra keywords passed are given as a list such as...
    #   ["--arg1", "val1", "--arg2", "val2",]
    # Using code from stack overflow, we parse that data into a dictionary.
    #   https://stackoverflow.com/questions/32944131/
    # If there's any issue parsing the data, we alert the user with a hint.
    try:
        kwargs_cleaned = {
            context.args[i][2:]: context.args[i + 1]
            for i in range(0, len(context.args), 2)
        }

    except:
        raise click.ClickException(
            "This workflow command appears to be improperly formatted. \n"
            "When giving workflow parameters, make sure you provide both the \n"
            "keyword and the value (such as `--example_kwarg example_value`)."
        )

    return kwargs_cleaned


@workflows.command()
def explore():
    """
    Let's you interactively view all available workflows and see the documentation
    on the one you select.
    """

    click.echo("\nGathering all available workflows...")
    from simmate.workflows.utilities import (
        get_workflow_types,
        get_list_of_workflows_by_type,
        get_workflow,
    )

    click.echo("\n\nWhat type of analysis are you interested in?")
    workflow_types = get_workflow_types()
    type_index = list_options(workflow_types)
    selected_type = workflow_types[type_index]

    # TODO: have the user select a calculator for this analysis. For now,
    # we are assuming VASP because those are our only workflows

    click.echo("\n\nWhat settings preset do you want to see the description for?")
    presets = get_list_of_workflows_by_type(selected_type, full_name=False)
    present_index = list_options(presets)
    selected_preset = presets[present_index]

    final_workflow_name = selected_type + ".vasp." + selected_preset
    # BUG: I assume vasp for now, but this should change when I add workflows
    # from new calculators.

    click.echo(f"\n\n===================== {final_workflow_name} =====================")

    # now we load this workflow and print the docstring.
    workflow = get_workflow(
        workflow_name=final_workflow_name,
        precheck_flow_exists=True,
        print_equivalent_import=True,
    )

    click.echo("Description:")
    click.echo(workflow.__doc__)

    click.echo("Parameters:")
    workflow.show_parameters()

    click.echo("==================================================================")


@workflows.command()
def list_all():
    """
    This lists off all available workflows.
    """

    click.echo("Gathering all available workflows...")

    from simmate.workflows.utilities import get_list_of_all_workflows

    click.echo("These are the workflows that have been registerd:")
    all_workflows = get_list_of_all_workflows()
    for i, workflow in enumerate(all_workflows):
        workflow_number = str(i + 1).zfill(2)
        click.echo(f"\t({workflow_number}) {workflow}")  # gives "(01) example-flow"


@workflows.command()
@click.argument("workflow_name")
def show_config(workflow_name):
    """
    If the workflow is a single task, the calculation's configuration settings
    are displayed. For example, a VASP workflow will show a dictionary that
    details how INCAR settings are selected.
    """

    click.echo("LOADING WORKFLOW...")
    from simmate.workflows.utilities import get_workflow

    workflow = get_workflow(
        workflow_name=workflow_name,
        precheck_flow_exists=True,
        print_equivalent_import=True,
    )

    click.echo("PRINTING WORKFLOW CONFIG...")

    # Not all workflows have a single config because some are NestWorkflows,
    # meaning they are made of multiple smaller workflows.
    if workflow.s3task:
        workflow.s3task.print_config()
    else:
        raise click.ClickException(
            "This is not a S3Task-based workflow. It is likely a NestedWorkflow, "
            "meaning it is made up of multiple smaller workflows. We have not "
            "added a show-config feature for these yet. "
        )


@workflows.command(
    context_settings=dict(
        ignore_unknown_options=True,
        allow_extra_args=True,
    )
)
@click.argument("workflow_name")
@click.pass_context
def setup_only(context, workflow_name):
    """
    If the workflow is a single task, the calculation is set up but not ran. This
    is useful when you just want the input files to view/edit.
    """

    click.echo("LOADING STRUCTURE AND WORKFLOW...")
    from simmate.toolkit import Structure
    from simmate.workflows.utilities import get_workflow

    workflow = get_workflow(
        workflow_name=workflow_name,
        precheck_flow_exists=True,
        print_equivalent_import=True,
    )
    kwargs_input = parse_parameters(context=context)
    kwargs_cleaned = workflow._deserialize_parameters(**kwargs_input)

    click.echo("WRITING INPUT FILES...")

    # if no folder was given, just name it after the workflow. We also replace
    # spaces with underscores
    from simmate.utilities import get_directory

    directory = kwargs_cleaned.get("directory", None)
    if not directory:
        directory = get_directory(f"{workflow.name_full}.SETUP-ONLY")
        kwargs_cleaned["directory"] = directory

    # Not all workflows have a single input because some are NestWorkflows,
    # meaning they are made of multiple smaller workflows.
    if workflow.s3task:
        workflow.s3task.setup(**kwargs_cleaned)
    else:
        raise click.ClickException(
            "This is not a S3Task-based workflow. It is likely a NestedWorkflow, "
            "meaning it is made up of multiple smaller workflows. We have not "
            "added a setup-only feature for these yet. "
        )

    click.echo(f"Done! Your input files are located in {directory}")


@workflows.command(
    context_settings=dict(
        ignore_unknown_options=True,
        allow_extra_args=True,
    )
)
@click.argument("workflow_name")
@click.pass_context
def run(context, workflow_name):
    """Runs a workflow using provided parameters"""

    click.echo("LOADING WORKFLOW & INPUT PARAMETERS...")

    from simmate.workflows.utilities import get_workflow

    workflow = get_workflow(
        workflow_name=workflow_name,
        precheck_flow_exists=True,
        print_equivalent_import=True,
    )
    kwargs_cleaned = parse_parameters(context=context)

    click.echo("RUNNING WORKFLOW...")

    result = workflow.run(**kwargs_cleaned)

    # Let the user know everything succeeded
    if result.is_completed():
        click.echo("Success! All results are also stored in your database.")


@workflows.command(
    context_settings=dict(
        ignore_unknown_options=True,
        allow_extra_args=True,
    )
)
@click.argument("workflow_name")
@click.pass_context
def run_cloud(context, workflow_name):
    """Submits a workflow to Prefect cloud"""

    click.echo("LOADING WORKFLOW & INPUT PARAMETERS...")

    from simmate.workflows.utilities import get_workflow

    workflow = get_workflow(
        workflow_name=workflow_name,
        precheck_flow_exists=True,
        print_equivalent_import=True,
    )
    kwargs_cleaned = parse_parameters(context=context)

    click.echo("RUNNING WORKFLOW...")

    result = workflow.run_cloud(**kwargs_cleaned)

    # Let the user know everything succeeded
    if result.is_completed():
        click.echo("Success! All results are also stored in your database.")


@workflows.command(
    context_settings=dict(
        ignore_unknown_options=True,
        allow_extra_args=True,
    )
)
@click.argument("filename", type=click.Path(exists=True))
def run_yaml(filename):
    """Runs a workflow where parameters are loaded from a yaml file."""

    click.echo("LOADING WORKFLOW & INPUT PARAMETERS...")

    import yaml

    with open(filename) as file:
        parameters = yaml.full_load(file)  # this is updated below

    from simmate.workflows.utilities import get_workflow

    workflow = get_workflow(
        # we pop the workflow name so that it is also removed from the rest of kwargs
        workflow_name=parameters.pop("workflow_name"),
        precheck_flow_exists=True,
        print_equivalent_import=True,
    )

    # -------------------------------------------------------------------------

    # SPECIAL CASE -- Running customized workflows from yaml.
    # !!! When prefect 2.0 is available, I will be able to use **kwargs as an
    # input for workflow parameters -- in which case, I can move this functionality
    # to load_input_and_register.

    # For customized workflows, we need to completely change the format
    # that we provide the parameters. Customized workflows expect parameters
    # broken into a dictionary of
    #   {"workflow_base": ..., "input_parameters":..., "updated_settings": ...}
    if "workflow_base" in parameters.keys():

        parameters["input_parameters"] = {}
        parameters["updated_settings"] = {}

        for key, update_values in list(parameters.items()):
            # Skip the base keys
            if key in [
                "workflow_base",
                "input_parameters",
                "updated_settings",
            ]:
                continue
            # if there is no prefix, then we have a normal input parameter
            elif not key.startswith("custom__"):
                parameters["input_parameters"][key] = parameters.pop(key)
            # Otherwise remove the prefix and add it to the custom settings.
            else:
                key_cleaned = key.removeprefix("custom__")
                parameters["updated_settings"][key_cleaned] = parameters.pop(key)

    # -------------------------------------------------------------------------

    click.echo("RUNNING WORKFLOW...")

    result = workflow.run(**parameters)

    # Let the user know everything succeeded
    if result.is_completed():
        click.echo("Success! All results are also stored in your database.")


# explicitly list functions so that pdoc doesn't skip them
__all__ = [
    "workflows",
    "explore",
    "list_all",
    "show_config",
    "setup_only",
    "run",
    "run_cloud",
    "run_yaml",
    "list_options",
    "parse_parameters",
]



# -*- coding: utf-8 -*-

"""

> :warning: This file is only for use by the Simmate team. Users should instead
access data via the load_remote_archive method.

This file is for pulling Materials Project data into the Simmate database. 
PyMatGen offers an easy way to do this in python -- the MPRester class. All you
need is [an API key from their site](https://materialsproject.org/open) and pymatgen
installed. For now, we only pull the mp-id, structure, and final energy.

"""

from django.db import transaction

from tqdm import tqdm
from pymatgen.ext.matproj import MPRester

from simmate.database.third_parties import MatprojStructure


@transaction.atomic
def load_all_structures(
    api_key: str,
    criteria: dict = {"task_id": {"$exists": True}},
    update_stabilities: bool = False,
):
    """
    Only use this function if you are part of the Simmate dev team!

    Loads all structures directly for the Material Project database into the
    local Simmate database.

    #### Parameters

    - `api_key`:
        Your Materials Project API key.
    - `criteria`:
        Filtering criteria for which structures to load. The default is all
        existing structures (137,885 as of 2022-01-16), which will take rouhghly
        15 min to complete (not including stabilities).
    - `update_stabilities`:
        Whether to run update_all_stabilities on the database table. Note this
        will add over an hour to this process. Default is True.
    """

    # Notes on filtering criteria for structures in the Materials Project:
    #
    # Catagories such as 'elements' that we can filter off of are listed here:
    #       https://github.com/materialsproject/mapidoc
    # Conditions such as $in or $exists that we filter based on are listed here:
    #       https://docs.mongodb.com/manual/reference/operator/query/
    #
    # As the simplest example, I only want one structure and I grab it by it's
    # mp-id in this code:
    #   criteria = {"task_id": mp_id}
    #
    # Here, we want all structures! Which is why we use:
    #   criteria = {"task_id": {"$exists": True}}
    #
    # This is an alternative criteria input that can be used for testing
    # criteria={
    #     "task_id": {
    #         "$exists": True,
    #         "$in": ["mp-" + str(n) for n in range(1, 10000)]},
    # }
    #

    # Connect to their database with personal API key
    mpr = MPRester(api_key)

    # For the filtered structures, this lists off which properties to grab.
    # All possible properties are listed here:
    #       https://github.com/materialsproject/mapidoc
    properties = [
        "material_id",
        "final_energy",
        "structure",
    ]

    # now make the query and grab everything from the Materials Project!
    # the output dictionary is given back within a list, where each entry is
    # a specific structure (so a single mp-id)
    # Note: this is a very large query, so make sure your computer has enough
    # memory (RAM >10GB) and a stable internet connection.
    data = mpr.query(criteria, properties)

    # Let's iterate through each structure and save it to the database
    # This also takes a while, so we use a progress bar
    # BUG: We init and update tqdm separately beacuse it conflicts with MPRester's bar
    progress_bar = tqdm(total=len(data), position=0)
    for entry in data:

        # update the progress bar
        progress_bar.update(1)

        # convert the data to a Simmate database object
        structure_db = MatprojStructure.from_toolkit(
            id=entry["material_id"],
            structure=entry["structure"],
            energy=entry["final_energy"],
        )

        # and save it to our database!
        structure_db.save()

    # once all structures are saved, let's update the Thermodynamic columns
    if update_stabilities:
        MatprojStructure.update_all_stabilities()



# -*- coding: utf-8 -*-

from simmate.workflow_engine import Workflow
from simmate.calculators.vasp.tasks.relaxation import (
    MatprojRelaxation as MPRelaxationTask,
)
from simmate.calculators.vasp.database.relaxation import (
    MatprojRelaxation as MPRelaxationResults,
)


class Relaxation__Vasp__Matproj(Workflow):
    s3task = MPRelaxationTask
    database_table = MPRelaxationResults
    description_doc_short = "uses Materials Project settings"



# -*- coding: utf-8 -*-

"""
Customized workflows involve modifying another workflow's core settings and
thus these are meant for testing.
"""

from simmate.calculators.vasp.workflows.customized import Customized__Vasp__UserConfig



# -*- coding: utf-8 -*-

from simmate.toolkit.transformations.base import Transformation


class SoftMutation(Transformation):

    # construct a dynamical matrix based on bond hardness and mutate site coords along the softest mode
    # https://uspex-team.org/static/file/USPEX-LargeComplexSystems-2010.pdf

    # This appears to be a means to order a disordered structure, which I can use other pymatgen methods for
    # from pymatgen.transformations.standard_transformations import OrderDisorderedStructureTransformation
    # from pymatgen.transformations.advanced_transformations import EnumerateStructureTransformation
    # After reading through the docs of these functions, it actually looks like "disordered" refers to structures with mixed occupancy sites (e.g. 50% Li / 50% Na)
    # so... I don't think these functions will help
    # the same goes for the reverse of these functions:
    # from pymatgen.transformations.advanced_transformations import DisorderOrderedTransformation

    # I have found that ASE has a version of this function written already
    # https://wiki.fysik.dtu.dk/ase/tutorials/ga/ga_bulk.html
    # https://gitlab.com/ase/ase/-/tree/master/ase/ga
    # from ase.ga.bulk_mutations import SoftMutation

    io_scale = "one_to_one"
    ninput = 1
    use_multiprocessing = False

    def __init__(
        self,
        composition,
        ratio_of_covalent_radii=0.1,
        bounds=[0.5, 2.0],
        use_tags=False,
        used_modes_file=None,
    ):  #!!! add all SoftMutation options or use **kwargs?

        # we can assume the user has ASE installed because it is a dependency of PyMatgen
        #!!! it looks like the ase.ga module is actively changing so version may introduce errors
        from ase.ga.soft_mutation import SoftMutation
        from ase.ga.utilities import closest_distances_generator

        # the closest_distances_generator is exactly the same as an element-dependent distance matrix
        # expect ASE puts this in dictionary form
        # the function requires a list of element integers
        element_ints = [element.number for element in composition]
        # the default of the ratio of covalent radii (0.1) is based on the ASE tutorial of this function
        element_distance_matrix = closest_distances_generator(
            element_ints, ratio_of_covalent_radii
        )

        # now make the SoftMutation mutator
        self.softmut = SoftMutation(
            blmin=element_distance_matrix,
            bounds=bounds,
            use_tags=use_tags,
            used_modes_file=used_modes_file,
        )

        # we also need to convert pymatgen Structures to ase Atoms below
        #!!! is it faster and more memory efficient to import below?
        from pymatgen.io.ase import AseAtomsAdaptor

        self.adaptor = AseAtomsAdaptor

    def apply_transformation(self, structure):

        ### CHECK FOR BUGS

        # This mutation is not possible for structures that have only one site
        if structure.num_sites == 1:
            # print('You cannot perform a soft mutation on structures that only have one site!! You can make this a supercell and try again though.')
            return False

        ### RUN

        # first I need to convert the structure to an ASE atoms object
        structure_ase = self.adaptor.get_atoms(structure)

        # their code searches for a atoms.info['confid'] to check for previously used modes -- it's unfortunately dependent on the entire ga structure
        # config stands for configuration, which I'm not actually using. So I set this to some random value which has no effect.
        structure_ase.info.update(
            {"confid": 0}
        )  #!!! I don't understand what this index represents and need to contact the dev

        # reset the used_modes to none
        #!!! I could also fill this if I want to try different modes until I have a unique structure
        #!!! the dictionary would be {0:[3,4,5...]} if I want to skip modes
        self.softmut.used_modes = {}

        #!!! Their code suggests the use of .get_new_individual() but I think .mutate() is what we'd like
        # new_structure_ase, label = self.softmut.get_new_individual([structure_ase])
        new_structure_ase = self.softmut.mutate(structure_ase)

        # if the mutation failed, None will be returned
        if new_structure_ase:
            # now convert back to a pymatgen object
            new_structure = self.adaptor.get_structure(new_structure_ase)
        # this indicates the mutation failed
        #!!! should I retry?
        else:
            new_structure = False

        return new_structure



# -*- coding: utf-8 -*-

import os

from simmate.conftest import copy_test_files, make_dummy_files

from simmate.toolkit.diffusion import MigrationImages

from simmate.calculators.vasp.inputs import Potcar
from simmate.calculators.vasp.tasks.nudged_elastic_band import MITNudgedElasticBand
from simmate.calculators.vasp.inputs.potcar_mappings import (
    PBE_ELEMENT_MAPPINGS_LOW_QUALITY,
)


def test_neb_setup(sample_structures, tmpdir, mocker):

    # To test this task we need to create images, which we do using I diffusion
    # in Y2CI2. We use [0] to grab the shortest path.
    structure = sample_structures["Y2CI2_mp-1206803_primitive"]
    images = MigrationImages.from_structure(structure, "I")[0]

    # estabilish filenames that we make and commonly reference
    incar_filename = os.path.join(tmpdir, "INCAR")
    potcar_filename = os.path.join(tmpdir, "POTCAR")
    # These files exist within a series of directories 00, 01,..., 05
    poscar_filenames = [
        os.path.join(tmpdir, str(n).zfill(2), "POSCAR") for n in range(5)
    ]

    # Because we won't have POTCARs accessible, we need to cover this function
    # call -- specifically have it pretend to make a file
    mocker.patch.object(
        Potcar,
        "to_file_from_type",
        return_value=make_dummy_files(potcar_filename),
    )

    # try to make input files in the tmpdir
    MITNudgedElasticBand.setup(
        migration_images=images,
        directory=tmpdir,
    )
    assert os.path.exists(incar_filename)
    assert os.path.exists(potcar_filename)
    assert all([os.path.exists(f) for f in poscar_filenames])
    Potcar.to_file_from_type.assert_called_with(
        structure.composition.elements,
        "PBE",
        potcar_filename,
        PBE_ELEMENT_MAPPINGS_LOW_QUALITY,
    )


def test_neb_workup(tmpdir):
    copy_test_files(
        tmpdir,
        test_directory=__file__,
        test_folder="nudged_elastic_band",
    )

    # init main task
    task = MITNudgedElasticBand()

    # estabilish filenames that we make and commonly reference
    summary_filename = os.path.join(tmpdir, "simmate_summary.yaml")
    plot_filename = os.path.join(tmpdir, "NEB_plot.jpeg")
    cif_filename = os.path.join(tmpdir, "path_relaxed_neb.cif")

    # run the full workup
    task.workup(tmpdir)
    assert os.path.exists(summary_filename)
    # assert os.path.exists(plot_filename)  # temporarily removed
    assert os.path.exists(cif_filename)



# -*- coding: utf-8 -*-

import os

from simmate.conftest import copy_test_files
from simmate.calculators.vasp.inputs import Incar
from simmate.calculators.vasp.error_handlers import IncorrectSmearing


def test_incorrect_smearing(tmpdir):
    copy_test_files(
        tmpdir,
        test_directory=__file__,
        test_folder="incorrect_smearing",
    )

    # we reference the files several spots below so we grab its path up front
    incar_filename = os.path.join(tmpdir, "INCAR")

    # init class with default settings
    error_handler = IncorrectSmearing()

    # Confirm an error IS NOT found
    error_handler.filename_to_check = "vasprun.does_not_exist"
    assert error_handler.check(tmpdir) == False

    # Confirm an error IS found
    error_handler.filename_to_check = "vasprun.xml"
    assert error_handler.check(tmpdir) == True

    # Make first attempt at fixing the error
    fix = error_handler.correct(tmpdir)
    assert fix == "switched ISMEAR to 2 amd SIGMA to 0.2"
    assert Incar.from_file(incar_filename)["ISMEAR"] == 2
    assert Incar.from_file(incar_filename)["SIGMA"] == 0.2



# -*- coding: utf-8 -*-

##############################################################################

import os
import pandas as pd
import itertools

##############################################################################


def loadWyckoffData():

    """
    This function loads the csv file containing all wyckoff site data into
    a pandas dataframe. One column that we need is not directly stored in
    the csv file though - this is site "Availablility", which can be inferred
    from the "Coordinates" column. Therefore, this function generates the
    "Availability" column before returning the pandas dataframe.
    """

    # Grab the directory of this current python file
    current_directory = os.path.dirname(__file__)

    # Set the full path to the data csv file
    datafile = current_directory + "/wyckoffdata.csv"

    # Load the csv file into a pandas dataframe
    data = pd.read_csv(datafile)

    # We need "Availibility" of each site as a column of the dataframe.
    # This can be found using the "Coordinates" column:
    # A wycloff site can only be used once (i.e. 0,0,0)
    # or an infinite number of times (i.e x,y,z).
    avail = []
    # Iterate through each site.
    for wy_site in data["Coordinates"].values:  # .values grabs column as numpy
        # The wy_site can be used infinite number of times if there
        # is a x,y, or z coordinate available.
        if any(i in wy_site for i in "xyz"):
            availability = float("inf")
        # The wy_site can only be used once if there
        # is no x,y, or z in the coordinates.
        else:
            availability = 1
        avail.append(availability)
    # We now have the final list and can add the new column to the dataframe.
    data["Availability"] = avail

    # return the data
    return data


##############################################################################


def findValidWyckoffCombos(stoich, spacegroup, wy_data=loadWyckoffData()):

    """
    Given a composition's stoichiometry (such as [4,4,12] for Mg4Si4O12) and
    a single spacegroup (1-230), this function will find all valid wyckoff
    combinations that can produce that stoichiometry.

    stoich = a list of integers representing the target stoichiometry
    spacegroup = an integer for the target spacegroup
    wy_data = pandas dataframe with wyckoff data (DON'T EDIT THIS!)

    NOTE: don't change the wy_data input! It is set in the header instead of
    running inside the function for time performance reasons. By putting it
    in the header, calling findValidWyckoffCombos() repetitively will not
    repetitively call loadWyckoffData() which yields a massive
    speed improvement.
    """

    # Grab all the wyckoff sites associated with the spacegroup given
    wy_sg = wy_data.query("SpaceGroup == @spacegroup")

    # This separate wy_sites into unique (MultiplicityPrimitive, Availability)
    # groups. This is useful for massive speed-up in the function as we can
    # find combos of these groups instead of all wy_sites. For example, all
    # wy_sites with Multiplicity = 2 and Availability = 2 will be treated as
    # one group when making combos then when that combo is used (in a
    # different function), it randomly grabs one wy_site from the group.
    wy_groups = wy_sg.groupby(["MultiplicityPrimitive", "Availability"]).groups

    # First, we need to find what the valid combinations are for each of the
    # individual elements. This code finds all combinations of wyckoffs sites.
    # Combinations can be up to nsites. So we go through all combinations
    # and see which ones meet the following criteria:
    #   1) total wyckoff sites used is less than nsites
    #   2) the total multiplicity of all wyckoff sites is equal to nsites
    #   3) no wyckoff site with an availability of 1 is used multiple times
    wy_combos = []
    for i, nsites in enumerate(stoich):
        # Don't search for wyckoff combinations if it's already been done for
        # a given nsite size. In [4,4,12] example, we run this code for
        # nsites=4 twice, which is unneccessary. Instead, we want to copy the
        # result of the first run, which is much faster. This IF statement
        # (and all code in it) is therefore a shortcut. The algorithm works
        # the same without this check, just slower.
        if nsites in stoich[:i]:  # if nsites value is earlier in stoich list
            # Find where in stoich that the nsites was done before
            i_match = stoich.index(nsites)
            # The index of the first occurence in stoich will match the
            # index in wy_combos. Copy the result of that wy_combo result over.
            wy_combos.append(wy_combos[i_match])
            # then skip to the next nsites
            continue
        # Make an empty list that we can store all the valid combinations in
        valid_group_combos = []
        # Try each combination size up to the value of nsite
        for s in range(1, nsites + 1):  # +1 because range stops at value prior
            # Find all combinations at the combo size of s
            combos = itertools.combinations_with_replacement(wy_groups, s)
            # For each combination, check that criteria are met
            for combo in combos:
                # Say the combo is valid until proven otherwise
                is_valid = True
                # The total multiplicity of all wyckoff sites must be equal to
                # nsites. Note, index 0 is "MultiplicityPrimitive".
                mults = sum([wy_group[0] for wy_group in combo])
                if mults != nsites:
                    # No need to check further, so jump to the next combo.
                    continue  # same as is_valid=False, but faster
                # No wy_site with an availability of 1 can used multiple times.
                # For example, site 0,0,0 cant be used more than once.
                for wy_group in combo:
                    # The 'uses' is total times that a given wy_group is used
                    uses = combo.count(wy_group)
                    # The 'avail_s' is the availabilitiy of a single wy_site
                    # in the wy_group, which is either 1 or infinite (inf).
                    avail_s = wy_group[1]
                    # If a single wy_site in the group has Availability = 1
                    if avail_s == 1:
                        # The total times that a wy_group is allowed to be
                        # used is the length of the group. For example, the
                        # group of 6 wy_sites that have Availability = 1 can
                        # be used a maximum of 6 times.
                        avail_g = wy_groups[wy_group].size
                    # If a single wy_site in the group has Availability = inf
                    else:
                        # The total times that this group can be used is inf.
                        # Note index 1 grabs the'inf' availability value.
                        avail_g = wy_group[1]
                    # Now check if the site is overused.
                    if uses > avail_g:
                        is_valid = False
                        break  # no need to check further in this combo
                # If the combo has passed all tests if it reached this point,
                # then add it to the valid list.
                if is_valid:
                    valid_group_combos.append(combo)
        wy_combos.append(valid_group_combos)

    # We now need to find the unique combinations of wy_combos accross all
    # elements. For example, we ensure that Mg and Si both can't use 0,0,0.
    # Make an empty list that we can store all the valid combinations in
    valid_combos = []
    #!!! The "*" uses "star expression, which I don't fully understand but it
    #!!! gives the desired output that I want.
    combos = itertools.product(*wy_combos)
    for combo in combos:
        # say the combo is valid until proven otherwise
        is_valid = True
        # The combo is invalid if a wycoff sites use is more than it's total
        # availibility. For example, the wyckoff site of (0,0,0) cannot be
        # used by both Mg and Si, whereas the wyckoff site of (0,0,z) can be
        # used by multiple elements.
        for wy_group in wy_groups:
            # The 'uses' is total times that a given wy_group is used
            uses = sum([element.count(wy_group) for element in combo])
            # The 'avail_s' is the availabilitiy of a single wy_site
            # in the wy_group, which is either 1 or infinite (inf).
            avail_s = wy_group[1]
            # If a single wy_site in the group has Availability = 1
            if avail_s == 1:
                # The total times that a wy_group is allowed to be
                # used is the length of the group. For example, the
                # group of 6 wy_sites that have Availability = 1 can
                # be used a maximum of 6 times.
                avail_g = wy_groups[wy_group].size
            # If a single wy_site in the group has Availability = inf
            else:
                # The total times that this group can be used is inf.
                # Note index 1 grabs the'inf' availability value.
                avail_g = wy_group[1]
            if uses > avail_g:
                is_valid = False
                break  # no need to check further in this combo
        # If the combo has passed all tests if it reached this point,
        # then add it to the valid list.
        if is_valid:
            valid_combos.append(combo)
    # We want to return the valid combinations of wyckoff groups because
    # groups can refer to a number of wy_sites (a,b,c, etc.)
    #!!! TO-DO: add a better explanation of what's being returned here
    return {"WyckoffGroups": wy_groups, "ValidCombinations": valid_combos}


##############################################################################


def findValidWyckoffCombosForListofSpacegroups(
    stoich, sg_include=range(1, 231), sg_exclude=[]
):

    """
    Find all wyckoff group combinations for specified list of spacegroups.
    This function calls the findValidWyckoffCombos function repeatedly, so
    really this is just a convience function.

    stoich = list of nsites for each element
        (i.e. Mg4Si4O12 has stoich = [4,4,12])
    sg_include = list of spacegroups that we are interested in.
        (default is all 230 spacegroups)
    sg_exclude = list of spacegroups that we should explicitly ignore
    """

    # Combine the inputs into a list of spacegroups we need to investigate
    sg_to_search = [sg for sg in sg_include if sg not in sg_exclude]

    # We save the outputs as a dictionary where the spacegroup (integer) is
    # the key to combo results.
    #!!! Consider changing this to a list where index=spacegroup-1
    combos_by_sg = {}
    for spacegroup in sg_to_search:
        combos = findValidWyckoffCombos(stoich, spacegroup)
        combos_by_sg.update({spacegroup: combos})

    return combos_by_sg


##############################################################################


def loadAsymmetricUnitData():

    """
    This function simply loads the csv file containing all asymmetric unit
    data into a pandas dataframe. It then just grabs the column we need and
    turns it into a numpy array (for performance reasons) even though we lose
    the spacegroup column - it's really just index+1 for each entry.
    """

    # grab the directory of this current python file
    current_directory = os.path.dirname(__file__)

    # set the full path to the data csv file
    datafile = current_directory + "/asymdata.csv"

    # load the csv file into a pandas dataframe
    data = pd.read_csv(datafile)

    # for speed, limit output to just a numpy array of the asym column
    data = data["asymmetricunit"].values

    # return the data
    return data


##############################################################################


def loadSpecifiedUnitData():

    """
    THIS IS JUST A COPY/PASTE OF loadAsymmetricUnitData() WITH A DIFFERENT
    COLUMN GRAB I NEED TO UPDATE THESE CSV FILES OR COMBINE THESE FUNCTIONS.
    """

    current_directory = os.path.dirname(__file__)
    datafile = current_directory + "/asymdata.csv"
    data = pd.read_csv(datafile)
    data = data["cellspec"].values
    return data


##############################################################################



# -*- coding: utf-8 -*-

import os
import shutil
from tempfile import TemporaryDirectory

from simmate.conftest import copy_test_files
from simmate.utilities.files import (
    get_directory,
    # make_archive,
    make_error_archive,
    archive_old_runs,
    empty_directory,
)


def test_get_directory(tmpdir):

    # create and delete a new directory
    new_directory = get_directory()
    assert "simmate-task-" in new_directory
    shutil.rmtree(new_directory)

    # get directory by name
    new_directory = get_directory(tmpdir)
    assert tmpdir == new_directory

    # Use a TemporaryDir instance
    tempdir2 = TemporaryDirectory()
    new_directory = get_directory(tempdir2)
    assert new_directory == tempdir2.name
    tempdir2.cleanup()


def test_make_archive(tmpdir):

    copy_test_files(
        tmpdir,
        test_directory=__file__,
        test_folder="to_archive",
    )

    archive_old_runs(tmpdir, time_cutoff=0)
    assert os.path.exists(os.path.join(tmpdir, "simmate-task-1.zip"))
    assert os.path.exists(os.path.join(tmpdir, "simmate-task-2.zip"))


def test_make_error_archive(tmpdir):

    copy_test_files(
        tmpdir,
        test_directory=__file__,
        test_folder="to_archive",
    )

    make_error_archive(tmpdir)
    assert os.path.exists(os.path.join(tmpdir, "simmate_attempt_01.zip"))

    make_error_archive(tmpdir)
    assert os.path.exists(os.path.join(tmpdir, "simmate_attempt_02.zip"))


def test_empty_directory(tmpdir):

    copy_test_files(
        tmpdir,
        test_directory=__file__,
        test_folder="to_archive",
    )

    empty_directory(tmpdir)
    assert not os.path.exists(os.path.join(tmpdir, "simmate-task-1"))
    assert not os.path.exists(os.path.join(tmpdir, "simmate-task-2"))



# -*- coding: utf-8 -*-

import os
import shutil

from simmate.workflow_engine import ErrorHandler
from simmate.calculators.vasp.inputs import Incar


class Zbrent(ErrorHandler):
    """
    Calculation is simply restarted using the most recent structure (CONTCAR)
    """

    is_monitor = True
    filename_to_check = "vasp.out"
    possible_error_messages = [
        "ZBRENT: fatal internal in",
        "ZBRENT: fatal error in bracketing",
    ]

    def correct(self, directory: str) -> str:

        # load the INCAR file to view the current settings
        incar_filename = os.path.join(directory, "INCAR")
        incar = Incar.from_file(incar_filename)

        # make the fix
        incar["IBRION"] = 1
        poscar_filename = os.path.join(directory, "POSCAR")
        contcar_filename = os.path.join(directory, "CONTCAR")
        shutil.copyfile(contcar_filename, poscar_filename)
        correction = "switched IBRION to 1 and copied the CONTCAR over to the POSCAR"

        # rewrite the INCAR with new settings
        incar.to_file(incar_filename)

        return correction



# -*- coding: utf-8 -*-

import os

from simmate.conftest import copy_test_files
from simmate.calculators.vasp.inputs import Incar
from simmate.calculators.vasp.error_handlers import Potim


def test_potim(tmpdir):
    copy_test_files(
        tmpdir,
        test_directory=__file__,
        test_folder="potim",
    )

    # we reference the files several spots below so we grab its path up front
    incar_filename = os.path.join(tmpdir, "INCAR")
    oszicar_filename = os.path.join(tmpdir, "OSZICAR")

    # init class with default settings
    error_handler = Potim()

    # Confirm an error IS found using test files
    assert error_handler.check(tmpdir) == True

    # Make first attempt at fixing the error
    fix = error_handler.correct(tmpdir)
    assert fix == "switched IBRION to 3 and SMASS to 0.75"
    incar = Incar.from_file(incar_filename)
    assert incar["IBRION"] == 3
    assert incar["SMASS"] == 0.75

    # Make second attempt at fixing the error
    fix = error_handler.correct(tmpdir)
    assert fix == "switched SYMPREC to 1e-8"
    assert Incar.from_file(incar_filename)["SYMPREC"] == 1e-8

    # Make third attempt at fixing the error
    fix = error_handler.correct(tmpdir)
    assert fix == "halved the POTIM from 0.01 to 0.005"
    assert Incar.from_file(incar_filename)["POTIM"] == 0.005

    # Confirm an error IS NOT found when there is no OSZICAR
    os.remove(oszicar_filename)
    assert error_handler.check(tmpdir) == False



# -*- coding: utf-8 -*-

"""
Want to make a website interface for your app? Then you'll need to fill this file
out! If you ever venture to this level, we strongly recommend you go through the
django tutorials first: https://docs.djangoproject.com/en/3.2/

COMING SOON: Simmate will automatically build out views for a workflow based
on what your database table looks like. Let us know if you are waiting on this
feature, so we can prioritize it!
"""

from django.shortcuts import render

from simmate.website.core_components.base_api_view import SimmateAPIViewSet

from .models import ExampleRelaxationTable

# Simple views (like a homepage) are made following django's normal format
def example_view(request):
    context = {}
    template = "path/to/my/template.html"
    return render(request, template, context)


# Advanced views that automatically generate a user interface and API for you
# use a class from Simmate!
class ExampleRelaxationViewSet(SimmateAPIViewSet):
    table = ExampleRelaxationTable
    # NOTE: default templates are still under development. For now, you can
    # reference our own templates to get started.
    template_list = "path/to/my/template1.html"
    template_retrieve = "path/to/my/template2.html"
